from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Mapping, Set

from .subs import subs

if TYPE_CHECKING:
    import numpy as np
    import sympy
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import sympy


def eval_expr(
    expr: sympy.Expr,
    datasize: int,
    datamap: Mapping[sympy.Expr, np.ndarray],
) -> np.ndarray:
    # NOTE: We avoid querying for free_symbols if we know none are expected
    fs = _free_symbols(expr) if datamap else set()

    if fs:
        ordered_symbols, fn = _lambdify_canonical(expr)
        data = [datamap[rv] for rv in ordered_symbols]
        return fn(*data)

    return np.full(datasize, float(expr.evalf()))  # pyright: ignore [reportGeneralTypeIssues]


@lru_cache(maxsize=256)
def _free_symbols(expr: sympy.Expr) -> Set[sympy.Expr]:
    return expr.free_symbols  # pyright: ignore [reportGeneralTypeIssues]


@lru_cache(maxsize=256)
def _lambdify_canonical(expr: sympy.Expr):
    fs = _free_symbols(expr)
    ordered_symbols = sorted(fs, key=str)
    # NOTE: Substitution allows to use cse. Otherwise weird things happen with
    # symbols that look like function eval (e.g. ETA(1), THETA(3), OMEGA(1,1)).
    ordered_substitutes = [sympy.Symbol(f'__tmp{i}') for i in range(len(ordered_symbols))]
    substituted_expr = subs(
        expr,
        dict(zip(ordered_symbols, ordered_substitutes)),
        simultaneous=True,
    )
    fn = sympy.lambdify(ordered_substitutes, substituted_expr, modules='numpy', cse=True)
    return ordered_symbols, fn



import altair
from altair.utils.execeval import eval_block


def pharmpy_theme():
    return {
        'config': {
            'axis': {
                'labelFontSize': 11,
                'titleFontSize': 13,
            },
            'legend': {
                'labelFontSize': 12,
                'titleFontSize': 13,
            },
        }
    }


altair.themes.register('pharmpy', pharmpy_theme)
altair.themes.enable('pharmpy')
altair.data_transformers.disable_max_rows()

__all__ = ('altair', 'eval_block')



# See issue #20
# extensions = ['altair.sphinxext.altairplot']
extensions = ['pharmpy.reporting.altairplot', 'jupyter_sphinx']  # TODO: Remove altairplot
html_sidebars = {
    '**': ['localtoc.html'],
}
html_static_path = ['.']
html_css_files = [
    'custom.css',
]
html_theme_options = {'body_max_width': 'none'}
html_title = 'Results'
master_doc = 'results'



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling.tmdd import DV_TYPES
from pharmpy.tools import summarize_modelfit_results
from pharmpy.tools.common import ToolResults, create_results, update_initial_estimates
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

from .drugmetabolite import create_drug_metabolite_models
from .pkpd import create_baseline_pd_model, create_pkpd_models
from .tmdd import create_qss_models, create_remaining_models

TYPES = frozenset(('pkpd', 'drug_metabolite', 'tmdd'))


def create_workflow(
    type: Literal[tuple(TYPES)],
    search_space: Optional[str] = None,
    b_init: Optional[Union[int, float]] = None,
    emax_init: Optional[Union[int, float]] = None,
    ec50_init: Optional[Union[int, float]] = None,
    met_init: Optional[Union[int, float]] = None,
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
    extra_model: Optional[Model] = None,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs >= 0.1)",
    extra_model_results: Optional[ModelfitResults] = None,
    dv_types: Optional[dict[Literal[DV_TYPES], int]] = None,
):
    """Run the structsearch tool. For more details, see :ref:`structsearch`.

    Parameters
    ----------
    type : str
        Type of model. Currently only 'drug_metabolite' and 'pkpd'
    search_space : str
        Search space to test
    b_init: float
        Initial estimate for the baseline for pkpd models. The default value is 0.1
    emax_init: float
        Initial estimate for E_MAX (for pkpd models only). The default value is 0.1
    ec50_init: float
        Initial estimate for EC_50 (for pkpd models only). The default value is 0.1
    met_init: float
        Initial estimate for MET (for pkpd models only). The default value is 0.1
    results : ModelfitResults
        Results for the start model
    model : Model
        Pharmpy start model
    extra_model : Model
        Optional extra Pharmpy model to use in TMDD structsearch
    extra_model_results : ModelfitResults
        Results for the extra model
    strictness : str or None
        Strictness criteria
    dv_types : dict
        Dictionary of DV types for TMDD models with multiple DVs

    Returns
    -------
    StructSearchResult
        structsearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import run_structsearch, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_structsearch(model_type='pkpd', results=results, model=model)   # doctest: +SKIP
    """

    wb = WorkflowBuilder(name="structsearch")
    if type == 'tmdd':
        start_task = Task(
            'run_tmdd',
            run_tmdd,
            model,
            results,
            extra_model,
            extra_model_results,
            strictness,
            dv_types,
        )
    elif type == 'pkpd':
        start_task = Task(
            'run_pkpd',
            run_pkpd,
            model,
            results,
            search_space,
            b_init,
            emax_init,
            ec50_init,
            met_init,
            strictness,
        )
    elif type == 'drug_metabolite':
        start_task = Task(
            'run_drug_metabolite', run_drug_metabolite, model, search_space, results, strictness
        )
    wb.add_task(start_task)
    return Workflow(wb)


def run_tmdd(context, model, results, extra_model, extra_model_results, strictness, dv_types):
    model = update_initial_estimates(model, results)
    model_entry = ModelEntry.create(model, modelfit_results=results)

    qss_candidate_models = create_qss_models(model, results.parameter_estimates, dv_types)
    qss_candidate_entries = [
        ModelEntry.create(m, modelfit_results=None, parent=model) for m in qss_candidate_models
    ]

    if extra_model is not None:
        extra_model = update_initial_estimates(extra_model, extra_model_results)
        extra_qss_candidate_models = create_qss_models(
            extra_model, extra_model_results.parameter_estimates, dv_types, index=9
        )
        extra_qss_candidate_entries = [
            ModelEntry.create(model, modelfit_results=None, parent=extra_model)
            for model in extra_qss_candidate_models
        ]
        qss_candidate_entries += extra_qss_candidate_entries

    wf = create_fit_workflow(qss_candidate_entries)
    wb = WorkflowBuilder(wf)
    task_results = Task('results', bundle_results)
    wb.add_task(task_results, predecessors=wf.output_tasks)
    qss_run_entries = call_workflow(Workflow(wb), 'results_QSS', context)

    ofvs = [
        model_entry.modelfit_results.ofv if model_entry.modelfit_results is not None else np.nan
        for model_entry in qss_run_entries
    ]
    minindex = ofvs.index(np.nanmin(ofvs))
    best_qss_entry = qss_run_entries[minindex]
    best_qss_entry = ModelEntry.create(
        best_qss_entry.model, modelfit_results=best_qss_entry.modelfit_results
    )

    models = create_remaining_models(
        model,
        best_qss_entry.modelfit_results.parameter_estimates,
        len(best_qss_entry.model.statements.ode_system.find_peripheral_compartments()),
        dv_types,
    )
    remaining_model_entries = [
        ModelEntry.create(model, modelfit_results=None, parent=best_qss_entry.model)
        for model in models
    ]

    wf2 = create_fit_workflow(remaining_model_entries)
    wb2 = WorkflowBuilder(wf2)
    task_results = Task('results', bundle_results)
    wb2.add_task(task_results, predecessors=wf2.output_tasks)
    run_model_entries = call_workflow(Workflow(wb2), 'results_remaining', context)

    summary_input = summarize_modelfit_results(results)
    summary_candidates = summarize_modelfit_results(
        [model_entry.modelfit_results for model_entry in qss_run_entries + run_model_entries]
    )

    return create_results(
        StructSearchResults,
        model_entry,
        best_qss_entry,
        list(run_model_entries),
        rank_type='bic',
        cutoff=None,
        summary_models=pd.concat([summary_input, summary_candidates], keys=[0, 1], names=['step']),
        strictness=strictness,
    )


def run_pkpd(
    context, input_model, results, search_space, b_init, emax_init, ec50_init, met_init, strictness
):
    model_entry = ModelEntry.create(input_model, modelfit_results=results)
    baseline_pd_model = create_baseline_pd_model(input_model, results.parameter_estimates, b_init)
    baseline_pd_model_entry = ModelEntry.create(baseline_pd_model, modelfit_results=None)

    wf = create_fit_workflow(baseline_pd_model_entry)
    wb = WorkflowBuilder(wf)
    task_results = Task('results2', bundle_results)
    wb.add_task(task_results, predecessors=wf.output_tasks)
    pd_baseline_fit = call_workflow(Workflow(wb), 'results_remaining', context)

    pkpd_models = create_pkpd_models(
        input_model,
        search_space,
        b_init,
        results.parameter_estimates,
        emax_init,
        ec50_init,
        met_init,
    )
    pkpd_model_entries = [
        ModelEntry.create(model, modelfit_results=None, parent=baseline_pd_model)
        for model in pkpd_models
    ]

    wf2 = create_fit_workflow(pkpd_model_entries)
    wb2 = WorkflowBuilder(wf2)
    task_results = Task('results2', bundle_results)
    wb2.add_task(task_results, predecessors=wf2.output_tasks)
    pkpd_models_fit = call_workflow(Workflow(wb2), 'results_remaining', context)

    summary_input = summarize_modelfit_results(results)
    summary_candidates = summarize_modelfit_results(
        [model_entry.modelfit_results for model_entry in pd_baseline_fit + pkpd_models_fit]
    )

    return create_results(
        StructSearchResults,
        model_entry,
        pd_baseline_fit[0],
        list(pkpd_models_fit),
        rank_type='bic',
        cutoff=None,
        summary_models=pd.concat([summary_input, summary_candidates], keys=[0, 1], names=["step"]),
        strictness=strictness,
    )


def run_drug_metabolite(context, model, search_space, results, strictness):
    model = update_initial_estimates(model, results)
    wb, candidate_model_tasks, base_model_description = create_drug_metabolite_models(
        model, results, search_space
    )

    task_results = Task(
        'Results',
        post_process_drug_metabolite,
        ModelEntry.create(model=model, modelfit_results=results),
        base_model_description,
        "bic",
        None,
        strictness,
    )

    wb.add_task(task_results, predecessors=candidate_model_tasks)
    results = call_workflow(Workflow(wb), "results_remaining", context)

    return results


def post_process_drug_metabolite(
    user_input_model_entry, base_model_description, rank_type, cutoff, strictness, *model_entries
):
    # NOTE : The base model is part of the model_entries but not the user_input_model
    res_models = []
    input_model_entry = None
    base_model_entry = None
    for model_entry in model_entries:
        model = model_entry.model
        if model.description == base_model_description:
            model_entry = ModelEntry.create(
                model=model_entry.model, parent=None, modelfit_results=model_entry.modelfit_results
            )
            input_model_entry = model_entry
            base_model_entry = model_entry
        else:
            res_models.append(model_entry)
    if not base_model_entry:
        # No base model found indicate user_input_model is base
        input_model_entry = user_input_model_entry
        base_model_entry = user_input_model_entry
    if not input_model_entry:
        raise ValueError('Error in workflow: No input model')

    if base_model_entry != user_input_model_entry:
        # Change parent model to base model instead of temporary model names or input model
        res_models = [
            me
            if me.parent.name
            not in ("TEMP", user_input_model_entry.model.name)  # TEMP name for drug-met models
            else ModelEntry.create(
                model=me.model, parent=base_model_entry.model, modelfit_results=me.modelfit_results
            )
            for me in res_models
        ]

    results_to_summarize = [user_input_model_entry.modelfit_results]

    if user_input_model_entry != base_model_entry:
        results_to_summarize.append(base_model_entry.modelfit_results)
    if res_models:
        results_to_summarize.extend(me.modelfit_results for me in res_models)

    summary_models = summarize_modelfit_results(results_to_summarize)
    summary_models['step'] = [0] + [1] * (len(summary_models) - 1)
    summary_models = summary_models.reset_index().set_index(['step', 'model'])

    return create_results(
        StructSearchResults,
        input_model_entry,
        base_model_entry,
        res_models,
        rank_type,
        cutoff,
        summary_models=summary_models,
        strictness=strictness,
    )


def bundle_results(*args):
    return args


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    type,
    strictness,
    model,
    dv_types,
    search_space,
    b_init,
    emax_init,
    ec50_init,
    met_init,
    extra_model,
    extra_model_results,
):
    if strictness is not None and "rse" in strictness.lower():
        if model.estimation_steps[-1].parameter_uncertainty_method is None:
            raise ValueError(
                'parameter_uncertainty_method not set for model, cannot calculate relative standard errors.'
            )

    if dv_types is not None:
        if not len(dv_types.values()) == len(set(dv_types.values())):
            raise ValueError('Values must be unique.')
        for key, value in dv_types.items():
            if key not in ['drug', 'drug_tot'] and value == 1:
                raise ValueError('Only drug can have DVID = 1. Please choose another DVID.')

    if type.lower() == 'tmdd':
        if search_space is not None:
            raise ValueError('Invalid argument "search_space" for TMDD models.')
        if any([b_init, emax_init, ec50_init, met_init]):
            raise ValueError(
                'Invalid arguments "b_init", "emax_init", "ec50_init" and "met_init" for TMDD models.'
            )
    elif type.lower() == 'pkpd':
        if extra_model is not None:
            raise ValueError('Invalid argument "extra_model" for PKPD models.')
        if extra_model_results is not None:
            raise ValueError('Invalid argument "extra_model_results" for PKPD models.')
        if dv_types is not None:
            raise ValueError('Invalid argument "dv_types" for PKPD models.')
    elif type.lower() == 'drug_metabolite':
        if any([b_init, emax_init, ec50_init, met_init]):
            raise ValueError(
                'Invalid arguments "b_init", "emax_init", "ec50_init" and "met_init" for drug metabolite models.'
            )
        if extra_model is not None:
            raise ValueError('Invalid argument "extra_model" for drug metabolite models.')
        if extra_model_results is not None:
            raise ValueError('Invalid argument "extra_model_results" for drug metabolite models.')
        if dv_types is not None:
            raise ValueError('Invalid argument "dv_types" for drug metabolite models.')


@dataclass(frozen=True)
class StructSearchResults(ToolResults):
    pass



from typing import List

from pharmpy.model import Model
from pharmpy.tools.mfl.helpers import funcs, structsearch_metabolite_features
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.tools.modelsearch.algorithms import exhaustive_stepwise
from pharmpy.workflows import ModelEntry, Task, WorkflowBuilder

from ..mfl.parse import parse as mfl_parse
from ..mfl.statement.feature.metabolite import Metabolite
from ..mfl.statement.feature.peripherals import Peripherals


def create_drug_metabolite_models(
    model: Model, results, search_space: str
) -> tuple[List[Model], Model]:
    # FIXME : Implement ModelFeatures when we can extract METABOLITE information

    mfl_statements = mfl_parse(search_space)
    metabolite_mfl_statements = [s for s in mfl_statements if isinstance(s, Metabolite)]
    metabolite_functions = funcs(model, metabolite_mfl_statements, structsearch_metabolite_features)
    peripheral_mfl_statements = [s for s in mfl_statements if isinstance(s, Peripherals)]
    peripheral_functions = funcs(model, peripheral_mfl_statements, structsearch_metabolite_features)

    # TODO: Update method for finding metabolite name
    if (
        model.statements.ode_system.find_compartment('METABOLITE')
        and len(metabolite_functions) != 0
    ):
        raise NotImplementedError(
            'Metabolite transformation on drug metabolite models is not yet possible.'
            ' Either remove METABOLITE transformations from search space or use another input model'
        )
    elif (
        not model.statements.ode_system.find_compartment('METABOLITE')
        and len(metabolite_functions) == 0
    ):
        raise ValueError(
            'Require at least one metabolite model type.'
            ' Try adding METABOLITE(BASIC) or METABOLITE(PSC) to search space'
        )

    if model.statements.ode_system.find_compartment('METABOLITE'):
        base_description = model.description
    else:
        base_description = determine_base_description(
            metabolite_mfl_statements, peripheral_mfl_statements
        )

    wb = WorkflowBuilder(name="drug_metabolite")

    start_task = Task("start", _start, model, results)
    wb.add_task(start_task)

    def apply_transformation(eff, f, model_entry):
        candidate_model = f(model_entry.model)
        candidate_model = candidate_model.replace(
            name="TEMP", description='_'.join(eff), modelfit_results=None
        )
        return ModelEntry.create(model=candidate_model, modelfit_results=None, parent=model)

    for eff, func in metabolite_functions.items():
        candidate_met_task = Task(str(eff), apply_transformation, eff, func)
        wb.add_task(candidate_met_task, predecessors=start_task)

    if len(peripheral_mfl_statements) == 0:
        candidate_model_tasks = []
        model_index = 1
        for out_task in wb.output_tasks:
            change_name_task = Task("number_run", change_name, model_index)
            model_index += 1
            wb.add_task(change_name_task, predecessors=[out_task])
            wf_fit = create_fit_workflow(n=1)
            wb.insert_workflow(wf_fit, predecessors=[change_name_task])

            candidate_model_tasks += wf_fit.output_tasks
    else:
        wb, candidate_model_tasks = exhaustive_stepwise(
            peripheral_functions, "no_add", wb, "structsearch"
        )

    return WorkflowBuilder(wb), candidate_model_tasks, base_description


def _start(model, results):
    return ModelEntry.create(model=model, modelfit_results=results)


def change_name(index, modelentry):
    return ModelEntry.create(
        model=modelentry.model.replace(name=f'structsearch_run{index}'),
        modelfit_results=modelentry.modelfit_results,
        parent=modelentry.parent,
    )


def determine_base_description(met_mfl, per_mfl):
    description = []
    if "BASIC" in [t.name for m in met_mfl for t in m.modes]:
        description.append("METABOLITE_BASIC")
    else:
        description.append("METABOLITE_PSC")
    if per_mfl:
        description.append(f"PERIPHERALS({min([c for p in per_mfl for c in p.counts])})")
    return ";".join(description)



from importlib import import_module
from types import ModuleType


class LazyImport(ModuleType):
    """Class that masquerades as a module and lazily loads it when accessed the first time

    The code for the class is taken from TensorFlow and is under the Apache 2.0 license

    This is needed for the CLI --version to be fast and in general for the
    library to load progressively.
    """

    def __init__(self, local_name, parent_module_globals, name, attr=None):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        self._attr = attr

        super(LazyImport, self).__init__(name)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = import_module(self.__name__)
        resolved = module if self._attr is None else getattr(module, self._attr)
        self._parent_module_globals[self._local_name] = resolved

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyImport, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(resolved.__dict__)
        return resolved

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)



import numpy as np
import pandas as pd
import pytest

from pharmpy.modeling import (
    create_rng,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    sample_parameters_uniformly,
)
from pharmpy.tools import read_modelfit_results


def test_create_rng():
    rng = create_rng(23)
    assert rng.standard_normal() == 0.5532605888887387

    rng = create_rng(23.0)
    assert rng.standard_normal() == 0.5532605888887387


def test_sample_parameters_uniformly(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    rng = create_rng(23)
    df = sample_parameters_uniformly(model, res.parameter_estimates, n=3, seed=rng)
    assert df['PTVCL'][0] == 0.004877674495376137


def test_sample_parameter_from_covariance_matrix(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    rng = np.random.default_rng(318)
    pe = res.parameter_estimates
    cm = res.covariance_matrix
    samples = sample_parameters_from_covariance_matrix(
        model,
        pe,
        cm,
        n=3,
        seed=rng,
    )
    correct = pd.DataFrame(
        {
            'PTVCL': [0.004489330033579095, 0.004866193232279955, 0.004619661658761273],
            'PTVV': [0.9720563045663096, 1.0217868717352445, 0.9662036500731115],
            'THETA_3': [0.19927467608338267, 0.237140948854298, 0.1979609848931148],
            'IVCL': [0.012012933626520568, 0.03859989956899462, 0.03228178862778379],
            'IVV': [0.03718187653238525, 0.036766142234483934, 0.02433717922068797],
            'SIGMA_1_1': [0.00962550646345379, 0.01311348785596405, 0.014054031573722888],
        }
    )
    pd.testing.assert_frame_equal(samples, correct, atol=1e-6)
    # Make cov matrix non-posdef
    cm2 = cm.copy()
    cm2['PTVCL']['PTVCL'] = -1
    with pytest.warns(UserWarning):
        sample_parameters_from_covariance_matrix(
            model,
            pe,
            cm2,
            n=1,
            force_posdef_covmatrix=True,
        )


def test_sample_individual_estimates(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    rng = np.random.default_rng(86)
    ie = res.individual_estimates
    iec = res.individual_estimates_covariance
    samples = sample_individual_estimates(model, ie, iec, seed=rng)
    assert len(samples) == 59 * 100
    assert list(samples.columns) == ['ETA_1', 'ETA_2']
    assert pytest.approx(samples.iloc[0]['ETA_1'], 1e-5) == 0.21179186940672637
    assert pytest.approx(samples.iloc[0]['ETA_2'], 1e-5) == -0.05771736555248238

    restricted = sample_individual_estimates(
        model, ie, iec, parameters=['ETA_2'], samples_per_id=1, seed=rng
    )
    assert len(restricted) == 59
    assert restricted.columns == ['ETA_2']
    assert pytest.approx(restricted.iloc[0]['ETA_2'], 1e-5) == 0.06399039578129821



import pandas as pd

from pharmpy.tools.bootstrap.results import calculate_results
from pharmpy.workflows.results import ModelfitResults, read_results


def test_bootstrap():
    orig = ModelfitResults(ofv=110, parameter_estimates=pd.Series({'TVCL': 0.75, 'TVV': 2.25}))
    res1 = ModelfitResults(ofv=100, parameter_estimates=pd.Series({'TVCL': 1.0, 'TVV': 2.0}))
    res2 = ModelfitResults(ofv=120, parameter_estimates=pd.Series({'TVCL': 1.5, 'TVV': 3.0}))

    class MyModel:
        pass

    res1_mod = MyModel()
    res2_mod = MyModel()

    boot = calculate_results([res1_mod, res2_mod], [res1, res2], original_results=orig)
    correct_statistics = pd.DataFrame(
        {
            'mean': [1.25, 2.5],
            'median': [1.25, 2.50],
            'bias': [0.5, 0.25],
            'stderr': [0.353553, 0.707107],
            'RSE': [0.2828424, 0.2828424],
        },
        index=['TVCL', 'TVV'],
    )

    pd.testing.assert_frame_equal(boot.parameter_statistics, correct_statistics)

    # json round trip
    json = boot.to_json()
    obj = read_results(json)
    pd.testing.assert_frame_equal(obj.parameter_statistics, correct_statistics)


def test_read_results(testdata):
    read_results(testdata / 'results/bootstrap_results.json')



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from pharmpy.deps import pandas as pd
from pharmpy.tools.common import ToolResults


@dataclass(frozen=True)
class COVSearchResults(ToolResults):
    steps: Optional[pd.DataFrame] = None
    ofv_summary: Optional[pd.DataFrame] = None
    candidate_summary: Optional[pd.DataFrame] = None



from typing import Union

import pharmpy.model
from pharmpy.deps import sympy
from pharmpy.internals.expr.subs import subs
from pharmpy.modeling import get_thetas


class res_error_term:
    def __init__(self, model, expr):
        self.model = model
        self.expr = sympy.expand(expr)

        self.only_piecewise = None
        self.is_only_piecewise()

        self.res = None
        self.res_alias = None
        self.add = error()
        self.prop = error()
        self.find_term()

    def create_res_alias(self):
        if self.res is not None:
            res_alias = set()
            for s in self.res.free_symbols:
                all_a = find_aliases(s, self.model)
                for a in all_a:
                    if a not in res_alias:
                        res_alias.add(a)
            self.res_alias = res_alias

    def dependencies(self):
        dependencies = set()
        if self.add.expr is not None:
            dependencies.update(self.add.dependencies)
        elif self.prop.expr is not None:
            dependencies.update(self.prop.dependencies)
        return dependencies

    def find_term(self):
        # NEED TO BE
        # VAR 1 : y = F + F*eps + eps
        # VAR 2 : y = F + F*theta*eps + theta*eps (EPS FIX 1)
        # VAR 3 : y = F + W*eps (W = sqrt(theta....)) (EPS FIX 1)
        # VAR 4 : y = F + W*eps (W = F*theta + theta) (EPS FIX 1)
        # ELSE conversion might be incorrect.
        terms = sympy.Add.make_args(self.expr)
        # Assert that it follows the above set of format rules
        assert len(terms) <= 3

        errors = []
        for term in terms:
            error_term = False
            full_term = full_expression(term, self.model)
            for factor in sympy.Mul.make_args(term):
                full_factor = full_expression(factor, self.model)
                all_symbols = full_factor.free_symbols.union(factor.free_symbols)
                for symbol in all_symbols:
                    if str(symbol) in self.model.random_variables.epsilons.names:
                        sigma = convert_eps_to_sigma(symbol, self.model)
                        if self.model.parameters[str(sigma)].init == 1.0:
                            if self.model.parameters[str(sigma)].fix:
                                term = term.subs(factor, 1)
                        if factor != symbol:
                            sigma_alias = factor
                        else:
                            sigma_alias = None
                        error_term = True

            if error_term:
                errors.append(
                    {
                        "term": term,
                        "full_term": full_term,
                        "sigma": sigma,
                        "sigma_alias": sigma_alias,
                    }
                )
            else:
                if self.res is None:
                    self.res = term
                    self.create_res_alias()
                else:
                    # FIXME: Should this be allowed?
                    self.res += term
                    self.create_res_alias()

        if self.res is None:
            print("No resulting term found")
            exit
        elif len(errors) > 2:
            print("Too many error terms found. Will try to translate either way.")

        for t in errors:
            prop = False
            ali_removed = False
            term = t["term"]
            full_term = t["full_term"]
            for symbol in full_term.free_symbols.union(term.free_symbols):
                for ali in find_aliases(symbol, self.model):
                    if ali in self.res_alias:
                        prop = True
                        # Remove the resulting symbol from the error term
                        term = convert_eps_to_sigma(term, self.model)
                        if ali in term.free_symbols:
                            term = term.subs(ali, 1)
                            ali_removed = True
            if prop is True:
                if not ali_removed:
                    term = term / self.res
                if self.prop.expr == 0:
                    self.prop = error(
                        self.model,
                        term,
                        t["sigma"],
                        sigma_alias=t["sigma_alias"],
                        prop=True,
                    )
                else:
                    self.prop.expr = self.prop.expr + term
            else:
                term = convert_eps_to_sigma(term, self.model)
                if self.add.expr == 0:
                    self.add = error(
                        self.model, term, t["sigma"], sigma_alias=t["sigma_alias"], add=True
                    )
                else:
                    self.add.expr = self.add.expr + term

    def is_only_piecewise(self):
        dv = list(self.model.dependent_variables.keys())[0]
        for s in reversed(self.model.statements.after_odes):
            if s.symbol == dv:
                if not s.expression.is_Piecewise:
                    self.only_piecewise = False
                    break

        if self.only_piecewise is None:
            self.only_piecewise = True

    def __str__(self):
        s = ""
        s += str(f"add : {self.add.expr}\n")
        s += str(f"add_sigma : {self.add.sigma}\n")
        s += str(f"prop : {self.prop.expr}\n")
        s += str(f"prop_sigma : {self.prop.sigma}\n")
        s += str(f"Only piecewise : {self.only_piecewise}\n")
        return s


class error:
    def __init__(self, model=None, expr=0, sigma=None, add=False, sigma_alias=None, prop=False):
        self.model = model
        self.expr = expr
        self.sigma = sigma
        self.sigma_alias = sigma_alias
        self.sigma_fix = self.is_sigma_fix()
        self.add = add
        self.prop = prop
        self.dependencies = set()
        self.check_dependecies()

    def is_sigma_fix(self):
        if self.model is not None:
            if self.model.parameters[str(self.sigma)].init == 1.0:
                if self.model.parameters[str(self.sigma)].fix:
                    return True
        else:
            return False

    def check_dependecies(self):
        if (
            self.model is not None
            and self.expr is not None
            and self.model.parameters[self.sigma].init == 1
            and self.sigma_fix
        ):
            accepted_symbols = set([self.sigma, self.sigma_alias])
            thetas = get_thetas(self.model).symbols
            accepted_symbols.update(thetas)
            etas = [sympy.Symbol(i) for i in self.model.random_variables.etas.names]
            accepted_symbols.update(etas)
            for symbol in self.expr.free_symbols:
                if not any([i in accepted_symbols for i in find_aliases(symbol, self.model)]):
                    if is_number(symbol, self.model):
                        accepted_symbols.update([symbol])
                    else:
                        self.dependencies.add(symbol)


def is_number(symbol: sympy.Expr, model: pharmpy.model.Model) -> bool:
    alias = find_aliases(symbol, model)
    for a in alias:
        if a not in model.random_variables.free_symbols:
            a_assign = model.statements.find_assignment(a)
            if a_assign.expression.is_number:
                return True
    return False


def full_expression(expression: sympy.Expr, model: pharmpy.model.Model) -> sympy.Expr:
    """
    Return the full expression of an expression (used for model statements)

    Parameters
    ----------
    expression : sympy.Expr
        Expression to be expanded.
    model : pharmpy.model.Model
        A pharmpy mode object with the expression as a statement.

    Returns
    -------
    expression : sympy.Expr
        The fully expanded expression

    """
    if len(model.statements.after_odes) == 0:
        statements = model.statements
    else:
        statements = model.statements.after_odes

    for statement in reversed(statements):
        expression = subs(expression, {statement.symbol: statement.expression}, simultaneous=True)
    return expression


def find_aliases(symbol: str, model: pharmpy.model, aliases: set = None) -> list:
    """
    Returns a list of all variable names that are the same as the inputed symbol

    Parameters
    ----------
    symbol : str
        The name of the variable to find aliases to.
    model : pharmpy.model
        A model by which the inputed symbol is related to.

    Returns
    -------
    aliases: list
        A list of aliases for the symbol.

    """
    if aliases is None:
        aliases = set([symbol])
    else:
        aliases.add(symbol)
    for expr in model.statements.after_odes:
        # If RES = ALI
        if symbol == expr.symbol and isinstance(expr.expression, sympy.Symbol):
            if expr.expression not in aliases:
                aliases.union(find_aliases(expr.expression, model, aliases))

        # If RES = PIECEWISE or PIECEWISE = RES
        if expr.expression.is_Piecewise:
            for e, c in expr.expression.args:
                if symbol == expr.symbol and isinstance(e, sympy.Symbol):
                    if e not in aliases:
                        aliases.union(find_aliases(e, model, aliases))
                elif symbol == e:
                    if expr.symbol not in aliases:
                        aliases.union(find_aliases(expr.symbol, model, aliases))

        # If ALI = RES
        if symbol == expr.expression:
            if expr.symbol not in aliases:
                aliases.union(find_aliases(expr.symbol, model, aliases))
    return aliases


def convert_eps_to_sigma(
    expr: Union[sympy.Symbol, sympy.Mul], model: pharmpy.model.Model
) -> Union[sympy.Symbol, sympy.Mul]:
    """
    Change the use of epsilon names to sigma names instead. Mostly used for
    converting NONMEM format to nlmxir2

    Parameters
    ----------
    expr : Union[sympy.Symbol,sympy.Mul]
        A sympy term to change a variable name in
    model : pharmpy.Model
        A pharmpy model object

    Returns
    -------
    Union[sympy.Symbol,sympy.Mul]
        Same expression as inputed, but with epsilon names changed to sigma.

    """
    # eps_to_sigma = {
    #    sympy.Symbol(eps.names[0]): sympy.Symbol(str(eps.variance))
    #    for eps in model.random_variables.epsilons
    # }
    eps_to_sigma = {}
    for dist in model.random_variables.epsilons:
        sigma = dist.variance
        if len(dist.names) == 1:
            eps_to_sigma[dist.names[0]] = sigma
        else:
            for row, col, eps in zip(range(sigma.rows), range(sigma.rows + 1), dist.names):
                eps_to_sigma[eps] = sigma[row, col]
    return expr.subs(eps_to_sigma)



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional, Set, Tuple, Union, overload

import pharmpy.internals.unicode as unicode
from pharmpy.internals.expr.assumptions import assume_all
from pharmpy.internals.expr.leaves import free_images, free_images_and_symbols
from pharmpy.internals.expr.matrix import is_zero_matrix
from pharmpy.internals.expr.ode import canonical_ode_rhs
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.immutable import Immutable, cache_method

if TYPE_CHECKING:
    import networkx as nx
    import sympy
else:
    from pharmpy.deps import networkx as nx
    from pharmpy.deps import sympy


class Statement(Immutable):
    """Abstract base class for all types of statements"""

    def __add__(self, other: Union[Statement, Statements, Sequence]) -> Statements:
        if isinstance(other, Statements):
            return Statements((self,) + other._statements)
        elif isinstance(other, Statement):
            return Statements((self, other))
        else:
            return Statements((self,) + tuple(other))

    def __radd__(self, other: Union[Statement, Sequence]) -> Statements:
        if isinstance(other, Statement):
            return Statements((other, self))
        else:
            return Statements(tuple(other) + (self,))

    @abstractmethod
    def subs(self, substitutions: Mapping[sympy.Basic, sympy.Basic]) -> Statement:
        pass

    @property
    @abstractmethod
    def free_symbols(self) -> Set[sympy.Expr]:
        pass

    @property
    @abstractmethod
    def rhs_symbols(self) -> Set[sympy.Expr]:
        pass


class Assignment(Statement):
    """Representation of variable assignment

    This class represents an assignment of an expression to a variable. Multiple assignments
    are combined together into a Statements object.

    Parameters
    ----------
    symbol : sympy.Symbol or str
        Symbol of statement
    expression : sympy.Expr
        Expression of assignment
    """

    def __init__(
        self, symbol: Union[sympy.Symbol, sympy.Function, sympy.Derivative], expression: sympy.Expr
    ):
        self._symbol = symbol
        self._expression = expression

    @classmethod
    def create(
        cls,
        symbol: Union[sympy.Symbol, sympy.Function, sympy.Derivative, str],
        expression: Union[sympy.Expr, int, float, str],
    ) -> Assignment:
        if isinstance(symbol, str):
            symbol = sympy.Symbol(symbol)
        if not (symbol.is_Symbol or symbol.is_Derivative or symbol.is_Function):
            raise TypeError("symbol of Assignment must be a Symbol or str representing a symbol")
        if isinstance(expression, str):
            expression = parse_expr(expression)
        elif isinstance(expression, float):
            expression = sympy.Float(expression)
        elif isinstance(expression, int):
            expression = sympy.Integer(expression)
        if expression.is_Piecewise:
            # To avoid nested piecewises
            expression = sympy.piecewise_fold(
                expression
            )  # pyright: ignore [reportGeneralTypeIssues]
        return cls(symbol, expression)  # pyright: ignore [reportGeneralTypeIssues]

    def replace(self, **kwargs):
        symbol = kwargs.get('symbol', self._symbol)
        expression = kwargs.get('expression', self._expression)
        return Assignment.create(symbol, expression)

    @property
    def symbol(self) -> Union[sympy.Symbol, sympy.Function, sympy.Derivative]:
        """Symbol of statement"""
        return self._symbol

    @property
    def expression(self) -> sympy.Expr:
        """Expression of assignment"""
        return self._expression

    def subs(self, substitutions: Mapping[sympy.Expr, sympy.Expr]) -> Assignment:
        """Substitute expressions or symbols in assignment

        Parameters
        ----------
        substitutions : dict
            old-new pairs

        Return
        ------
        Assignment
            Updated assignment object

        Examples
        --------
        >>> from pharmpy.model import Assignment
        >>> a = Assignment.create('CL', 'POP_CL + ETA_CL')
        >>> a
        CL = ETA_CL + POP_CL
        >>> b = a.subs({'ETA_CL' : 'ETA_CL * WGT'})
        >>> b
        CL = ETA_CL⋅WGT + POP_CL

        """
        symbol = subs(self.symbol, substitutions, simultaneous=True)
        assert (
            isinstance(symbol, sympy.Symbol)
            or isinstance(symbol, sympy.Derivative)
            or isinstance(symbol, sympy.Function)
        )
        expression = subs(self.expression, substitutions, simultaneous=True)
        if expression.is_Piecewise:
            # To avoid nested piecewises
            expression = sympy.piecewise_fold(expression)
        return Assignment(symbol, expression)

    @property
    def free_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the assignment

        Note that the left hand side symbol will be in the set

        Examples
        --------
        >>> from pharmpy.model import Assignment
        >>> a = Assignment.create('CL', 'POP_CL + ETA_CL')
        >>> a.free_symbols      # doctest: +SKIP
        {CL, ETA_CL, POP_CL}

        """
        symbols = {self.symbol}
        symbols |= self.expression.free_symbols
        return symbols  # pyright: ignore [reportGeneralTypeIssues]

    @property
    def rhs_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the right hand side expression

        Examples
        --------
        >>> from pharmpy.model import Assignment
        >>> a = Assignment.create('CL', 'POP_CL + ETA_CL')
        >>> a.rhs_symbols      # doctest: +SKIP
        {ETA_CL, POP_CL}

        """
        prefuncs = self._expression.atoms(sympy.Function)
        from sympy.core.function import AppliedUndef

        # Allow applied undefined functions
        funcs = {f for f in prefuncs if isinstance(f, AppliedUndef)}
        symbols = self._expression.free_symbols
        return funcs | symbols

    def __eq__(self, other):
        if hash(self) != hash(other):
            return False
        return (
            isinstance(other, Assignment)
            and self.symbol == other.symbol
            and self.expression == other.expression
        )

    @cache_method
    def __hash__(self):
        return hash((self.symbol, self.expression))

    def to_dict(self) -> dict[str, Any]:
        return {
            'class': 'Assignment',
            'symbol': sympy.srepr(self.symbol),
            'expression': sympy.srepr(self.expression),
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Assignment:
        return cls(
            symbol=sympy.parse_expr(d['symbol']), expression=sympy.parse_expr(d['expression'])
        )

    def __repr__(self):
        expression = sympy.pretty(self.expression)
        lines = [line.rstrip() for line in expression.split('\n')]
        definition = f'{sympy.pretty(self.symbol)} = '
        s = ''
        for line in lines:
            if line == lines[-1]:
                s += definition + line + '\n'
            else:
                s += len(definition) * ' ' + line + '\n'
        return s.rstrip()

    def _repr_latex_(self):
        sym = sympy.latex(self.symbol)
        expr = sympy.latex(self.expression, mul_symbol='dot')
        return f'${sym} = {expr}$'


class CompartmentBase(Immutable):
    pass


class Output(CompartmentBase):
    def __new__(cls):
        # Singleton class
        if not hasattr(cls, 'instance'):
            cls.instance = super(Output, cls).__new__(cls)
        return cls.instance

    def __repr__(self):
        return "Output()"

    def __hash__(self):
        return 5267

    def to_dict(self):
        return {'class': 'Output'}

    @classmethod
    def from_dict(cls, d):
        return cls.instance

    def __eq__(self, other):
        return self is other


output = Output()


class CompartmentalSystemBuilder:
    """Builder for CompartmentalSystem"""

    def __init__(self, cs: Optional[CompartmentalSystem] = None):
        if cs:
            self._g = cs._g.copy()
        else:
            self._g = nx.DiGraph()
            self._g.add_node(output)  # Single output "compartment"

    def add_compartment(self, compartment: Compartment) -> None:
        """Add compartment to system

        The compartment will be added without any flows to other compartments.
        Use the add_flow method to add flows to and from the newly added compartment.

        Parameters
        ----------
        compartment : Compartment
            Compartment to add

        Examples
        --------
        >>> from pharmpy.model import CompartmentalSystemBuilder
        >>> cb = CompartmentalSystemBuilder()
        >>> central = cb.add_compartment("CENTRAL")
        """
        self._g.add_node(compartment)

    def remove_compartment(self, compartment: Compartment) -> None:
        """Remove compartment from system

        Parameters
        ----------
        compartment : Compartment
            Compartment object to remove from system

        Examples
        --------
        >>> from pharmpy.model import CompartmentalSystemBuilder
        >>> cb = CompartmentalSystemBuilder()
        >>> central = Compartment.create("CENTRAL")
        >>> cb.add_compartment(central)
        >>> cb.remove_compartment(central)
        """
        self._g.remove_node(compartment)

    def add_flow(
        self, source: Compartment, destination: CompartmentBase, rate: Union[sympy.Basic, str]
    ) -> None:
        """Add flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment
        rate : Expression
            Symbolic rate of flow

        Examples
        --------
        >>> from pharmpy.model import Compartment, CompartmentalSystemBuilder
        >>> cb = CompartmentalSystemBuilder()
        >>> depot = Compartment.create("DEPOT")
        >>> cb.add_compartment(depot)
        >>> central = Compartment.create("CENTRAL")
        >>> cb.add_compartment(central)
        >>> cb.add_flow(depot, central, "KA")
        """
        self._g.add_edge(source, destination, rate=parse_expr(rate))

    def remove_flow(self, source: Compartment, destination: Compartment) -> None:
        """Remove flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment

        Examples
        --------
        >>> from pharmpy.model import CompartmentalSystemBuilder
        >>> cb = CompartmentalSystemBuilder()
        >>> depot = Compartment.create("DEPOT")
        >>> cb.add_compartment(depot)
        >>> central = Compartment.create("CENTRAL")
        >>> cb.add_compartment(central)
        >>> cb.add_flow(depot, central, "KA")
        >>> cb.remove_flow(depot, central)
        """
        self._g.remove_edge(source, destination)

    def move_dose(
        self,
        source: Compartment,
        destination: Compartment,
        admid: Optional[int] = None,
        replace: bool = True,
    ) -> Tuple[Compartment, Compartment]:
        """Move a dose input from one compartment to another

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment
        admid : int
            Move dose with specified admid, move all if None. Default is None.
        replace : bool
            Replace dose of destination or add to. Default to True

        """
        if admid:
            new_source_dose = tuple([d for d in source.doses if d.admid != admid])
            new_dest_dose = tuple([d for d in source.doses if d.admid == admid])
        else:
            new_source_dose = tuple()
            new_dest_dose = source.doses
        new_source = source.replace(doses=new_source_dose)
        if replace:
            new_dest = destination.replace(doses=source.doses)
        else:
            new_dest = destination.replace(doses=destination.doses + new_dest_dose)
        mapping = {source: new_source, destination: new_dest}
        nx.relabel_nodes(self._g, mapping, copy=False)
        return new_source, new_dest

    def set_dose(
        self,
        compartment: Compartment,
        dose: Optional[tuple[Dose, ...]],
        replace: bool = True,
        admid: Optional[int] = None,
    ) -> Compartment:
        """Set dose of compartment, replacing the previous by default.

        Set replace to False to instead add the dose to existing ones.

        Given an admid, the given dose will replace the corresponding dose
        based on that admid. This will ignore the 'replace' argument.

        Parameters
        ----------
        compartment : Compartment
            Compartment for which to change dose
        dose : Dose
            New dose
        replace : bool
            Replace existing doses or not. Default is True
        admid : Optional[int]
            Option to replace doses with specified admid.

        Returns
        -------
        Compartment
            The new updated compartment
        """
        if dose is None:
            dose = tuple()
        elif isinstance(dose, Dose):
            dose = (dose,)

        elif admid:
            replace = True
            old_dose = tuple([d for d in compartment.doses if d.admid != admid])
            dose = old_dose + dose

        if replace or not compartment.doses:
            new_comp = compartment.replace(doses=dose)
        else:
            new_comp = compartment.replace(doses=compartment.doses + dose)
        mapping = {compartment: new_comp}
        nx.relabel_nodes(self._g, mapping, copy=False)
        return new_comp

    def set_lag_time(self, compartment: Compartment, lag_time: sympy.Expr) -> Compartment:
        """Set lag time of compartment

        Parameters
        ----------
        compartment : Compartment
            Compartment for which to change lag time
        lag_time : expr
            New lag time

        Returns
        -------
        Compartment
            The new updated compartment
        """
        new_comp = compartment.replace(lag_time=lag_time)
        mapping = {compartment: new_comp}
        nx.relabel_nodes(self._g, mapping, copy=False)
        return new_comp

    def set_bioavailability(
        self, compartment: Compartment, bioavailability: sympy.Expr
    ) -> Compartment:
        """Set bioavailability of compartment

        Parameters
        ----------
        compartment : Compartment
            Compartment for which to change bioavailability
        bioavailability : expr
            New bioavailability

        Returns
        -------
        Compartment
            The new updated compartment
        """
        new_comp = compartment.replace(bioavailability=bioavailability)
        mapping = {compartment: new_comp}
        nx.relabel_nodes(self._g, mapping, copy=False)
        return new_comp

    def set_input(self, compartment: Compartment, input: sympy.Expr) -> Compartment:
        """Set zero order input of compartment

        Parameters
        ----------
        compartment : Compartment
            Compartment for which to change zero order input
        input : expr
            New input

        Returns
        -------
        Compartment
            The new updated compartment
        """
        new_comp = compartment.replace(input=input)
        mapping = {compartment: new_comp}
        nx.relabel_nodes(self._g, mapping, copy=False)
        return new_comp

    def find_compartment(self, name: str) -> Optional[Compartment]:
        for comp in self._g.nodes:
            if not isinstance(comp, Output) and comp.name == name:
                return comp


def _is_positive(expr: sympy.Basic) -> bool:
    return (
        sympy.ask(
            sympy.Q.positive(expr), assume_all(sympy.Q.positive, free_images_and_symbols(expr))
        )
        is True
    )


def _comps(graph):
    return {comp for comp in graph.nodes if not isinstance(comp, Output)}


def to_compartmental_system(names, eqs: Sequence[sympy.Eq]) -> CompartmentalSystem:
    """Convert an list of odes to a compartmental system

    names : func to compartment name map
    """

    cb = CompartmentalSystemBuilder()
    compartments = {}
    concentrations = set()
    for eq in eqs:
        A = eq.lhs.args[0]
        concentrations.add(A)
        name = names[A.func]
        comp = Compartment.create(name)
        cb.add_compartment(comp)
        compartments[name] = comp

    neweqs = list(eqs)  # Remaining flows

    for eq in eqs:
        checked_terms = set()
        for comp_func in concentrations.intersection(free_images(eq.rhs)):
            dep = eq.rhs.as_independent(comp_func, as_Add=True)[1]
            terms = sympy.Add.make_args(dep.expand())
            for term in terms:
                if term not in checked_terms:
                    checked_terms.add(term)
                    from_comp = None
                    to_comp = None
                    if len(concentrations.intersection(free_images(term))) >= 2:
                        # This means second order absorption -> find matching term
                        # to determine flow
                        if _is_positive(term):
                            for second_comp in concentrations.intersection(free_images(term)):
                                for eq_2 in eqs:
                                    if eq_2.lhs.args[0].name == second_comp.name:
                                        if -term in sympy.Add.make_args(eq_2.rhs.expand()):
                                            from_comp = compartments[names[second_comp.func]]
                                            to_comp = compartments[names[eq.lhs.args[0].func]]
                    else:
                        # Find matching term to determine if flow is between
                        # compartments or not
                        if _is_positive(term):
                            for eq_2 in eqs:
                                if -term in sympy.Add.make_args(eq_2.rhs.expand()):
                                    from_comp = compartments[names[eq_2.lhs.args[0].func]]
                                    to_comp = compartments[names[eq.lhs.args[0].func]]

                    if from_comp is not None and to_comp is not None:
                        # FIXME: Get current flow from builder instead?
                        cs = CompartmentalSystem(cb)
                        current_flow = cs.get_flow(from_comp, to_comp)

                        if isinstance(current_flow, sympy.core.numbers.Zero):
                            cb.add_flow(from_comp, to_comp, term / comp_func)
                        else:
                            new_flow = term / comp_func + current_flow
                            cb.add_flow(from_comp, to_comp, new_flow)

                        for i, neweq in enumerate(neweqs):
                            if neweq.lhs.args[0].name == eq.lhs.args[0].name:
                                neweqs[i] = sympy.Eq(neweq.lhs, sympy.expand(neweq.rhs - term))
                            elif neweq.lhs.args[0].name == comp_func.name:
                                neweqs[i] = sympy.Eq(neweq.lhs, sympy.expand(neweq.rhs + term))
    for eq in neweqs:
        if eq.rhs != 0:
            i = sympy.Integer(0)
            o = sympy.Integer(0)
            for term in sympy.Add.make_args(eq.rhs):
                if _is_positive(term):
                    i = i + term
                else:
                    o = o + term
            comp_func = eq.lhs.args[0]
            from_comp = compartments[names[comp_func.func]]
            if o != 0:
                cb.add_flow(from_comp, output, -o / comp_func)
            if i != 0:
                cb.set_input(from_comp, i)
    cs = CompartmentalSystem(cb)
    return cs


class CompartmentalSystem(Statement):
    """System of ODEs descibed as a compartmental system

    Examples
    --------
    >>> from pharmpy.model import Bolus, Compartment, output
    >>> from pharmpy.model import CompartmentalSystemBuilder, CompartmentalSystem
    >>> cb = CompartmentalSystemBuilder()
    >>> dose = Bolus.create("AMT")
    >>> central = Compartment.create("CENTRAL", doses=(dose,))
    >>> cb.add_compartment(central)
    >>> peripheral = Compartment.create("PERIPHERAL")
    >>> cb.add_compartment(peripheral)
    >>> cb.add_flow(central, peripheral, "K12")
    >>> cb.add_flow(peripheral, central, "K21")
    >>> cb.add_flow(central, output, "CL / V")
    >>> CompartmentalSystem(cb)        # doctest: +SKIP
    Bolus(AMT)
                     ┌──────────┐
                     │PERIPHERAL│
                     └──────────┘
                      ↑        │
                     K12      K21
                      │        ↓
    ┌───────┐      ┌──────────┐      ┌──────────┐
    │CENTRAL│──K12→│PERIPHERAL│──K21→│ CENTRAL  │──CL/V→
    └───────┘      └──────────┘      └──────────┘
    """

    def __init__(
        self,
        builder: Optional[CompartmentalSystemBuilder] = None,
        graph=None,
        t: sympy.Symbol = sympy.Symbol('t'),
    ):
        if builder is not None:
            self._g = nx.freeze(builder._g.copy())
        elif graph is not None:
            self._g = graph
        self._t = t

    @classmethod
    def create(
        cls,
        builder: Optional[CompartmentalSystemBuilder] = None,
        graph=None,
        eqs=None,
        t: sympy.Symbol = sympy.Symbol('t'),
    ) -> CompartmentalSystem:
        numargs = (
            (0 if builder is None else 1) + (0 if graph is None else 1) + (0 if eqs is None else 1)
        )
        if numargs != 1:
            raise ValueError(
                "Need exactly one of builder, graph or eqs to create a CompartmentalSystem"
            )
        if eqs is not None:
            pass
        t = sympy.sympify(t)
        return cls(builder=builder, graph=graph, t=t)

    def replace(self, **kwargs) -> CompartmentalSystem:
        t = kwargs.get('t', self._t)
        builder = kwargs.get('builder', None)
        if builder is None:
            g = kwargs.get('graph', self._g)
        else:
            g = None
        return CompartmentalSystem.create(builder=builder, graph=g, t=t)

    @property
    def t(self) -> sympy.Symbol:
        """Independent variable of CompartmentalSystem"""
        return self._t

    @property
    def eqs(self) -> Tuple[sympy.Eq, ...]:
        """Tuple of equations"""
        amount_funcs = sympy.Matrix(list(self.amounts))
        derivatives = sympy.Matrix([sympy.Derivative(fn, self.t) for fn in amount_funcs])
        inputs = self.zero_order_inputs
        a = self.compartmental_matrix @ amount_funcs + inputs
        eqs = [sympy.Eq(lhs, canonical_ode_rhs(rhs)) for lhs, rhs in zip(derivatives, a)]
        return tuple(eqs)

    @property
    def free_symbols(self) -> set[sympy.Basic]:
        """Get set of all free symbols in the compartmental system

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.free_symbols  # doctest: +SKIP
        {AMT, CL, V, t}
        """
        free = {sympy.Symbol('t')}
        for _, _, rate in self._g.edges.data('rate'):
            free |= rate.free_symbols
        for node in _comps(self._g):
            free |= node.free_symbols
        return free

    @property
    def rhs_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the right hand side expressions

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.rhs_symbols   # doctest: +SKIP
        {AMT, CL, V, t}
        """
        return self.free_symbols  # This works currently

    def subs(self, substitutions: Mapping[sympy.Basic, sympy.Basic]) -> CompartmentalSystem:
        """Substitute expressions or symbols in ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.subs({'AMT': 'DOSE'})
        Bolus(DOSE, admid=1) → CENTRAL
        ┌───────┐
        │CENTRAL│──CL/V→
        └───────┘
        """
        cb = CompartmentalSystemBuilder(self)
        for u, v, rate in cb._g.edges.data('rate'):
            assert isinstance(rate, sympy.Basic)
            rate_sub = subs(rate, substitutions, simultaneous=True)
            cb._g.edges[u, v]['rate'] = rate_sub
        mapping = {comp: comp.subs(substitutions) for comp in _comps(self._g)}
        nx.relabel_nodes(cb._g, mapping, copy=False)
        return CompartmentalSystem(cb)

    def atoms(self, cls):
        """Get set of all symbolic atoms of some kind

        For more information see
        https://docs.sympy.org/latest/modules/core.html#sympy.core.basic.Basic.atoms

        Parameters
        ----------
        cls : type
            Type of atoms to find

        Returns
        -------
        set
            Set of symbolic atoms
        """
        atoms = set()
        for _, _, rate in self._g.edges.data('rate'):
            atoms |= rate.atoms(cls)
        return atoms

    def __eq__(self, other):
        return (
            isinstance(other, CompartmentalSystem)
            and self._t == other._t
            and nx.to_dict_of_dicts(self._g) == nx.to_dict_of_dicts(other._g)
            and self.dosing_compartments == other.dosing_compartments
        )

    def __hash__(self):
        return hash((self._t, self._g))

    def to_dict(self) -> dict[str, Any]:
        comps = [comp for comp in self._g.nodes]
        comps_dicts = tuple(comp.to_dict() for comp in comps)

        edges = []
        for from_comp, to_comp, rate in self._g.edges.data('rate'):
            from_n = comps.index(from_comp)
            to_n = comps.index(to_comp)
            edge = (from_n, to_n, sympy.srepr(rate))
            edges.append(edge)

        d = {
            'class': 'CompartmentalSystem',
            'compartments': comps_dicts,
            'rates': edges,
            't': sympy.srepr(self._t),
        }
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> CompartmentalSystem:
        cb = CompartmentalSystemBuilder()
        comps = []
        for comp_dict in d['compartments']:
            if comp_dict['class'] == 'Output':
                compartment = output
            else:
                compartment = Compartment.from_dict(comp_dict)
                cb.add_compartment(compartment)
            comps.append(compartment)

        for from_n, to_n, rate in d['rates']:
            from_comp = comps[from_n]
            to_comp = comps[to_n]
            cb.add_flow(from_comp, to_comp, sympy.parse_expr(rate))

        return cls(cb, t=sympy.parse_expr(d['t']))

    def get_flow(self, source: CompartmentBase, destination: CompartmentBase) -> sympy.Expr:
        """Get the rate of flow between two compartments

        Parameters
        ----------
        source : Compartment
            Source compartment
        destination : Compartment
            Destination compartment

        Returns
        -------
        Expression
            Symbolic rate

        Examples
        --------
        >>> from pharmpy.model import CompartmentalSystem, Compartment
        >>> cb = CompartmentalSystemBuilder()
        >>> depot = Compartment.create("DEPOT")
        >>> cb.add_compartment(depot)
        >>> central = Compartment.create("CENTRAL")
        >>> cb.add_compartment(central)
        >>> cb.add_flow(depot, central, "KA")
        >>> odes = CompartmentalSystem(cb)
        >>> odes.get_flow(depot, central)
        KA
        >>> odes.get_flow(central, depot)
        0
        """
        try:
            rate = self._g.edges[source, destination]['rate']
        except KeyError:
            rate = sympy.Integer(0)
        return rate

    def get_compartment_outflows(
        self, compartment: Union[str, CompartmentBase]
    ) -> list[tuple[CompartmentBase, sympy.Expr]]:
        """Get list of all flows going out from a compartment

        Parameters
        ----------
        compartment : Compartment or str
            Get outflows for this compartment

        Returns
        -------
        list
            Pairs of compartments and symbolic rates

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_compartment_outflows("CENTRAL")
        [(Output(), CL/V)]
        """
        compartment = self.find_compartment_or_raise(compartment)
        flows = []
        for node in self._g.successors(compartment):
            flow = self.get_flow(compartment, node)
            flows.append((node, flow))
        return flows

    def get_compartment_inflows(
        self, compartment: Union[CompartmentBase, str]
    ) -> list[tuple[Compartment, sympy.Expr]]:
        """Get list of all flows going in to a compartment

        Parameters
        ----------
        compartment : Compartment or str
            Get inflows to this compartment

        Returns
        -------
        list
            Pairs of compartments and symbolic rates

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_compartment_inflows(output)
        [(Compartment(CENTRAL, amount=A_CENTRAL(t), doses=Bolus(AMT, admid=1)), CL/V)]
        """
        if isinstance(compartment, str):
            destination = self.find_compartment(compartment)
            if destination is None:
                raise ValueError(f"Cannot find compartment {compartment}")
        else:
            destination = compartment
        flows = []
        for node in self._g.predecessors(destination):
            flow = self.get_flow(node, destination)
            flows.append((node, flow))
        return flows

    def get_bidirectionals(self, compartment: Union[CompartmentBase, str]) -> list[Compartment]:
        """Get list of all compartments with bidirectional flow from/to a compartment

        Parameters
        ----------
        compartment : Compartment or str
            Compartment of interest

        Returns
        -------
        list
            Compartments with bidirectional flow

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> central = model.statements.ode_system.central_compartment
        >>> model.statements.ode_system.get_bidirectionals(central)
        []
        """
        compartment = self.find_compartment_or_raise(compartment)
        comps = []
        for node in self._g.predecessors(compartment):
            if self._g.has_edge(compartment, node):
                comps.append(node)
        return comps

    def find_compartment(self, name: str) -> Optional[Compartment]:
        """Find a compartment using its name

        Parameters
        ----------
        name : str
            Name of compartment to find

        Returns
        -------
        Compartment
            Compartment named name or None if not found

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> central = model.statements.ode_system.find_compartment("CENTRAL")
        >>> central
        Compartment(CENTRAL, amount=A_CENTRAL(t), doses=Bolus(AMT, admid=1))
        """
        for comp in _comps(self._g):
            if comp.name == name:
                return comp
        else:
            return None

    def find_compartment_or_raise(self, comp: Union[str, CompartmentBase]) -> Compartment:
        if isinstance(comp, CompartmentBase):
            return comp
        found_comp = self.find_compartment(comp)
        if found_comp is None:
            raise ValueError(f"No compartment named {comp}")
        return found_comp

    def get_n_connected(self, comp: Compartment) -> int:
        """Get the number of compartments connected to a compartment

        Parameters
        ----------
        comp : Compartment
            The compartment

        Returns
        -------
        int
            Number of compartments connected to comp

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.get_n_connected("CENTRAL")
        0
        """
        out_comps = {c for c, _ in self.get_compartment_outflows(comp)}
        in_comps = {c for c, _ in self.get_compartment_inflows(comp)}
        return len((out_comps | in_comps) - {output})

    @property
    def dosing_compartments(self) -> tuple[Compartment, ...]:
        """The dosing compartment(s)

        A dosing compartment is a compartment that receives an input dose. Multiple
        dose compartments are supported. The order of dose compartments is defined
        to put the central compartment last.

        Returns
        -------
        tuple
            A tuple of dose compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.dosing_compartments
        (Compartment(CENTRAL, amount=A_CENTRAL(t), doses=Bolus(AMT, admid=1)),)
        """
        dosing_comps = tuple()
        for node in _comps(self._g):
            if node.doses:
                if node.name != self.central_compartment.name:
                    if len(dosing_comps) >= 2:
                        dosing_comps = dosing_comps[:-1] + (node,) + dosing_comps[-1:]
                    else:
                        dosing_comps = (node,) + dosing_comps
                else:
                    dosing_comps = dosing_comps + (node,)

        if len(dosing_comps) != 0:
            return dosing_comps

        raise ValueError('No dosing compartment exists')

    @property
    def central_compartment(self) -> Compartment:
        """The central compartment

        The central compartment is defined to be the compartment that has an outward flow
        to the output compartment. Only one central compartment is supported.

        Returns
        -------
        Compartment
            Central compartment

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.central_compartment
        Compartment(CENTRAL, amount=A_CENTRAL(t), doses=Bolus(AMT, admid=1))
        """
        try:
            # E.g. TMDD models have more than one output
            # FIXME: Relying heavily on compartment naming for drug metabolite
            central = list(self._g.predecessors(output))[-1]
            if central.name in ["METABOLITE", "EFFECT", "COMPLEX"]:
                central = self.find_compartment("CENTRAL")
                if central is None:
                    raise ValueError('Cannot find central compartment')
        except IndexError:
            raise ValueError('Cannot find central compartment')
        return central

    def find_peripheral_compartments(self, name: Optional[str] = None) -> list[Compartment]:
        """Find perihperal compartments

        A peripheral compartment is defined as having one flow to the central compartment and
        one flow from the central compartment.

        If name is set, peripheral compartments connected to the compartment
        with the associated name is returned.

        Returns
        -------
        list of compartments
            Peripheral compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.find_peripheral_compartments()
        []
        """
        if name is not None:
            central = self.find_compartment(name)
            if central is None:
                raise ValueError(f"{name} is not a name of a connected compartment")
        else:
            central = self.central_compartment
        oneout = {node for node, out_degree in self._g.out_degree() if out_degree == 1}
        onein = {node for node, in_degree in self._g.in_degree() if in_degree == 1}
        cout = {comp for comp in oneout if self.get_flow(comp, central) != 0}
        cin = {comp for comp in onein if self.get_flow(central, comp) != 0}
        peripherals = list(cout & cin)
        # Return in deterministic order
        peripherals = sorted(peripherals, key=lambda comp: comp.name)
        return peripherals

    def find_transit_compartments(self, statements: Statements) -> list[Compartment]:
        """Find all transit compartments

        Transit compartments are a chain of compartments with the same out rate starting from
        the dose compartment. Because one single transit compartment cannot be distinguished
        from one depot compartment such compartment will be defined to be a depot and not
        a transit compartment.

        Returns
        -------
        list of compartments
            Transit compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.find_transit_compartments(model.statements)
        []
        """
        transits = []
        comp = self.dosing_compartments[0]
        if len(self.get_compartment_inflows(comp)) != 0:
            return transits
        outflows = self.get_compartment_outflows(comp)
        if len(outflows) != 1:
            return transits
        transits.append(comp)
        comp, rate = outflows[0]
        rate = statements.before_odes.full_expression(rate)
        while True:
            if len(self.get_compartment_inflows(comp)) != 1:
                break
            outflows = self.get_compartment_outflows(comp)
            if len(outflows) != 1:
                break
            next_comp, next_rate = outflows[0]
            next_rate = statements.before_odes.full_expression(next_rate)
            if rate != next_rate:
                break
            transits.append(comp)
            comp = next_comp
        # Special case of one transit directly into central is not defined as a transit
        # Also not central itself
        central = self.central_compartment
        if len(transits) == 1 and (
            self.get_flow(transits[0], central) != 0 or transits[0] == central
        ):
            return []
        else:
            return transits

    def find_depot(self, statements: Statements) -> Compartment:
        """Find the depot compartment

        The depot compartment is defined to be the compartment that only has out flow to the
        central compartment, but no flow from the central compartment.

        For drug metabolite models however, it is possible to have outflow with
        unidirectional flow to both the central and metabolite compartment. In this case,
        the central compartment is found based on name.

        Returns
        -------
        Compartment
            Depot compartment

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> model = load_example_model("pheno")
        >>> model = set_first_order_absorption(model)
        >>> model.statements.ode_system.find_depot(model.statements)
        Compartment(DEPOT, amount=A_DEPOT(t), doses=Bolus(AMT, admid=1))
        """
        transits = self.find_transit_compartments(statements)
        depot = self._find_depot()
        if depot in transits:
            depot = None
        return depot

    def _find_depot(self):
        central = self.central_compartment
        depot = None
        for to_central, _ in self.get_compartment_inflows(central):
            outflows = self.get_compartment_outflows(to_central)
            if len(outflows) == 1 or len(outflows) == 2:
                if len(outflows) == 2:
                    # FIXME: Use another method than compartment name
                    metabolite = self.find_compartment("METABOLITE")
                    if metabolite is None or not self.get_flow(to_central, metabolite):
                        break
                inflows = self.get_compartment_inflows(to_central)
                for in_comp, _ in inflows:
                    if in_comp == central:
                        break
                else:
                    depot = to_central
                    break
        return depot

    @property
    def compartmental_matrix(self) -> sympy.Matrix:
        """Compartmental matrix of the compartmental system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model, set_first_order_absorption
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.compartmental_matrix)
        ⎡-CL ⎤
        ⎢────⎥
        ⎣ V  ⎦
        """
        nodes = self._order_compartments()
        size = len(nodes)
        f = sympy.zeros(size)
        for i in range(0, size):
            from_comp = nodes[i]
            diagsum = 0
            for j in range(0, size):
                to_comp = nodes[j]
                rate = self.get_flow(from_comp, to_comp)
                if i != j:
                    f[j, i] = rate
                diagsum += rate
            outrate = self.get_flow(from_comp, output)
            f[i, i] = -outrate - diagsum
        return f

    @property
    def amounts(self) -> sympy.Matrix:
        """Column vector of amounts for all compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.amounts)
        [A_CENTRAL(t)]
        """
        ordered_cmts = self._order_compartments()
        amts = [cmt.amount for cmt in ordered_cmts]
        return sympy.Matrix(amts)

    @property
    def compartment_names(self) -> list[str]:
        """Names of all compartments

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system.compartment_names
        ['CENTRAL']
        """
        ordered_cmts = self._order_compartments()
        names = [cmt.name for cmt in ordered_cmts]
        return names

    def _order_compartments(self):
        """Return list of all compartments in canonical order"""
        try:
            dosecmt = self.dosing_compartments[0]
        except ValueError:
            # Fallback for cases where no dose is available (yet)
            return list(_comps(self._g))
        # Order compartments

        def sortfunc(x):
            a = list(x)
            if output in a:
                a.remove(output)
            a = sorted(a, key=lambda x: x.name)
            return iter(a)

        nodes = list(nx.bfs_tree(self._g, dosecmt, sort_neighbors=sortfunc))
        remaining = set(self._g.nodes) - {output} - set(nodes)
        while remaining:  # Disjoint graph
            comp = sorted(remaining, key=lambda x: x.name)[0]
            # Start with a compartment having a zero order input
            for c in remaining:
                if comp is None and c.input != 0:
                    comp = c
            ordered_remaining = list(nx.bfs_tree(self._g, comp, sort_neighbors=sortfunc))
            cleaned_ordered_remaining = []
            for c in ordered_remaining:
                if c not in set(nodes):
                    cleaned_ordered_remaining.append(c)
            nodes += cleaned_ordered_remaining
            remaining = set(self._g.nodes) - {output} - set(nodes)
        return nodes

    @property
    def zero_order_inputs(self) -> sympy.Matrix:
        """Vector of all zero order inputs to each compartment

        Example
        -------
        >>> from pharmpy.modeling import load_example_model, set_zero_order_absorption
        >>> import sympy
        >>> model = load_example_model("pheno")
        >>> sympy.pprint(model.statements.ode_system.zero_order_inputs)
        [0]

        """
        inputs = [node.input for node in self._order_compartments()]
        return sympy.Matrix(inputs)

    def __len__(self):
        """The number of compartments"""
        return len(self._g.nodes) - 1

    def _repr_html_(self):
        # Use Unicode art for now. There should be ways of drawing networkx
        s = str(self)
        return f'<pre>{s}</pre>'

    def __repr__(self):
        def lag_string(comp):
            return f"\ntlag={comp.lag_time}" if comp.lag_time != 0 else ""

        def f_string(comp):
            return f"\nF={comp.bioavailability}" if comp.bioavailability != 1 else ""

        def comp_string(comp):
            return comp.name + lag_string(comp) + f_string(comp)

        current = self.dosing_compartments[0]
        comp_height = 0
        comp_width = 0

        while True:
            bidirects = self.get_bidirectionals(current)
            outflows = self.get_compartment_outflows(current)
            comp_height = max(comp_height, len(bidirects) + 1)
            comp_width += 1
            for comp, rate in outflows:
                if comp not in bidirects and comp != output:
                    current = comp
                    break
            else:
                break

        noutput = len(self.get_compartment_inflows(output))
        have_zo_input = int(not is_zero_matrix(self.zero_order_inputs))
        comp_nrows = comp_height * 2 - 1 + (1 if noutput > 1 else 0)
        nrows = comp_nrows + have_zo_input
        ncols = comp_width * 2
        grid = unicode.Grid(nrows, ncols)

        current = self.dosing_compartments[0]
        col = 0
        if comp_nrows == 1 or comp_nrows == 2:
            main_row = 0 + have_zo_input
        else:
            main_row = 2 + have_zo_input

        if have_zo_input:
            # Assuming a fully linear layout
            for i, zo in enumerate(self.zero_order_inputs):
                if zo != 0:
                    try:
                        grid.set(0, i * 2, unicode.VerticalArrow(str(zo)))
                    except IndexError:
                        warnings.warn(
                            """The ODE system cannot be printed. Try statements.before_odes
                            and statements.after_odes instead."""
                        )
        while True:
            bidirects = self.get_bidirectionals(current)
            outflows = self.get_compartment_outflows(current)
            comp_box = unicode.Box(comp_string(current))
            grid.set(main_row, col, comp_box)

            if bidirects:
                grid.set(0, col, unicode.Box(comp_string(bidirects[0])))
                grid.set(
                    1,
                    col,
                    unicode.DualVerticalArrows(
                        str(self.get_flow(current, bidirects[0])),
                        str(self.get_flow(bidirects[0], current)),
                    ),
                )
            if len(bidirects) > 1:
                grid.set(4, col, unicode.Box(comp_string(bidirects[1])))
                grid.set(
                    3,
                    col,
                    unicode.DualVerticalArrows(
                        str(self.get_flow(bidirects[1], current)),
                        str(self.get_flow(current, bidirects[1])),
                    ),
                )

            for comp, rate in outflows:
                if comp not in bidirects and comp != output:
                    next_comp = comp
                    break
            else:
                next_comp = None

            if next_comp:
                rate = self.get_flow(current, next_comp)
            else:
                rate = self.get_flow(current, output)
            arrow = unicode.Arrow(str(rate))
            grid.set(main_row, col + 1, arrow)
            if next_comp and noutput > 1:
                rate = self.get_flow(current, output)
                arrow = unicode.VerticalArrow(str(rate))
                grid.set(main_row + 1, col, arrow)
            col += 2
            current = next_comp
            if not current:
                break

        all_doses = ""
        for dose_comp in self.dosing_compartments:
            for dose in dose_comp.doses:
                all_doses += f'{str(dose)} → {dose_comp.name} \n'
        s = all_doses + str(grid).rstrip()
        return s


class Dose(ABC):
    """Abstract base class for different types of doses"""

    def __init__(self, admid: int):
        self._admid = admid

    @property
    def admid(self) -> int:
        """Administration ID of dose"""
        return self._admid

    @abstractmethod
    def subs(self, substitutions) -> Dose:
        ...

    @property
    @abstractmethod
    def free_symbols(self) -> set[sympy.Expr]:
        ...

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        ...


class Bolus(Dose, Immutable):
    """A Bolus dose

    Parameters
    ----------
    amount : symbol
        Symbolic amount of dose
    admid : int
        Administration ID

    Examples
    --------
    >>> from pharmpy.model import Bolus
    >>> dose = Bolus.create("AMT")
    >>> dose
    Bolus(AMT, admid=1)
    """

    def __init__(self, amount: sympy.Expr, admid: int = 1):
        self._amount = amount
        self._admid = admid

    @classmethod
    def create(cls, amount: Union[sympy.Expr, str], admid: int = 1) -> Bolus:
        return cls(parse_expr(amount), admid=admid)

    def replace(self, **kwargs) -> Bolus:
        amount = kwargs.get("amount", self._amount)
        admid = kwargs.get("admid", self._admid)
        return Bolus.create(amount=amount, admid=admid)

    @property
    def amount(self) -> sympy.Expr:
        """Symbolic amount of dose"""
        return self._amount

    @property
    def free_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the dose

        Examples
        --------
        >>> from pharmpy.model import Bolus
        >>> dose = Bolus.create("AMT")
        >>> dose.free_symbols
        {AMT}
        """
        return {self.amount}

    def subs(self, substitutions: Mapping[sympy.Expr, sympy.Expr]) -> Bolus:
        """Substitute expressions or symbols in dose

        Parameters
        ----------
        substitutions : dict
            Dictionary of from, to pairs

        Examples
        --------
        >>> from pharmpy.model import Bolus
        >>> dose = Bolus.create("AMT")
        >>> dose.subs({'AMT': 'DOSE'})
        Bolus(DOSE, admid=1)
        """
        return Bolus(subs(self.amount, substitutions, simultaneous=True), admid=self._admid)

    def __eq__(self, other):
        return (
            isinstance(other, Bolus)
            and self._amount == other._amount
            and self._admid == other._admid
        )

    def __hash__(self):
        return hash((self._amount, self._admid))

    def to_dict(self) -> dict[str, Any]:
        return {'class': 'Bolus', 'amount': sympy.srepr(self._amount), 'admid': self._admid}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Bolus:
        return cls(amount=sympy.parse_expr(d['amount']), admid=d['admid'])

    def __repr__(self):
        return f'Bolus({self.amount}, admid={self._admid})'


class Infusion(Dose, Immutable):
    """An infusion dose

    Parameters
    ----------
    amount : expression
        Symbolic amount of dose
    admid : int
        Administration ID
    rate : expression
        Symbolic rate. Mutually exclusive with duration
    duration : expression
        Symbolic duration. Mutually excluseive with rate

    Examples
    --------
    >>> from pharmpy.model import Infusion
    >>> dose = Infusion("AMT", duration="D1")
    >>> dose
    Infusion(AMT, admid=1, duration=D1)
    >>> dose = Infusion("AMT", rate="R1")
    >>> dose
    Infusion(AMT, admid=1, rate=R1)
    """

    def __init__(
        self,
        amount: sympy.Expr,
        admid: int = 1,
        rate: Optional[sympy.Expr] = None,
        duration: Optional[sympy.Expr] = None,
    ):
        self._amount = amount
        self._admid = admid
        self._rate = rate
        self._duration = duration

    @classmethod
    def create(
        cls,
        amount: Union[str, sympy.Expr],
        admid: int = 1,
        rate: Optional[Union[str, sympy.Expr]] = None,
        duration: Optional[Union[str, sympy.Expr]] = None,
    ) -> Infusion:
        if rate is None and duration is None:
            raise ValueError('Need rate or duration for Infusion')
        if rate is not None and duration is not None:
            raise ValueError('Cannot have both rate and duration for Infusion')
        if rate is not None:
            rate = parse_expr(rate)
        if duration is not None:
            duration = parse_expr(duration)
        return cls(parse_expr(amount), admid=admid, rate=rate, duration=duration)

    def replace(self, **kwargs) -> Infusion:
        amount = kwargs.get("amount", self._amount)
        admid = kwargs.get("admid", self._admid)
        rate = kwargs.get("rate", self._rate)
        duration = kwargs.get("duration", self._duration)
        return Infusion.create(amount=amount, admid=admid, rate=rate, duration=duration)

    @property
    def amount(self) -> sympy.Expr:
        """Symbolic amount of dose"""
        return self._amount

    @property
    def rate(self) -> Optional[sympy.Expr]:
        """Symbolic rate

        Mutually exclusive with duration.
        """
        return self._rate

    @property
    def duration(self) -> Optional[sympy.Expr]:
        """Symbolc duration

        Mutually exclusive with rate.
        """
        return self._duration

    @property
    def free_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the dose

        Examples
        --------
        >>> from pharmpy.model import Infusion
        >>> dose = Infusion.create("AMT", rate="RATE")
        >>> dose.free_symbols   # doctest: +SKIP
        {AMT, RATE}
        """
        if self.rate is not None:
            symbs = self.rate.free_symbols
        else:
            assert self.duration is not None
            symbs = self.duration.free_symbols
        return symbs | self.amount.free_symbols

    def subs(self, substitutions: Mapping[sympy.Basic, sympy.Basic]) -> Infusion:
        """Substitute expressions or symbols in dose

        Parameters
        ----------
        substitutions : dict
            Dictionary of from, to pairs

        Examples
        --------
        >>> from pharmpy.model import Infusion
        >>> dose = Infusion.create("AMT", duration="DUR")
        >>> dose.subs({'DUR': 'D1'})
        Infusion(AMT, admid=1, duration=D1)
        """
        amount = subs(self.amount, substitutions, simultaneous=True)
        if self.rate is not None:
            rate = subs(self.rate, substitutions, simultaneous=True)
            duration = None
        else:
            rate = None
            duration = subs(self.duration, substitutions, simultaneous=True)
        return Infusion(amount, admid=self._admid, rate=rate, duration=duration)

    def __eq__(self, other):
        return (
            isinstance(other, Infusion)
            and self._admid == other._admid
            and self._rate == other._rate
            and self._duration == other._duration
            and self._amount == other._amount
        )

    def __hash__(self):
        return hash((self._admid, self._rate, self._duration, self._amount))

    def to_dict(self) -> dict[str, Any]:
        return {
            'class': 'Infusion',
            'amount': sympy.srepr(self._amount),
            'rate': sympy.srepr(self._rate),
            'duration': sympy.srepr(self._duration),
            'admid': self._admid,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Infusion:
        return cls(
            amount=sympy.parse_expr(d['amount']),
            admid=d['admid'],
            rate=sympy.parse_expr(d['rate']),
            duration=sympy.parse_expr(d['duration']),
        )

    def __repr__(self):
        if self.rate is not None:
            arg = f'rate={self.rate}'
        else:
            arg = f'duration={self.duration}'
        return f'Infusion({self.amount}, admid={self._admid}, {arg})'


class Compartment(CompartmentBase):
    """Compartment for a compartmental system

    Parameters
    ----------
    name : str
        Compartment name
    doses : tuple(Dose)
        Dose object for dose into this compartment. Default None for no dose.
    input : Expression
        Expression for other inputs to the compartment
    lag_time : Expression
        Lag time for doses entering this compartment. Default 0
    bioavailability : Expression
        Bioavailability fraction for doses entering this compartment. Default 1

    Examples
    --------
    >>> from pharmpy.model import Bolus, Compartment
    >>> comp = Compartment.create("CENTRAL")
    >>> comp
    Compartment(CENTRAL, amount=A_CENTRAL(t))
    >>> comp = Compartment.create("DEPOT", lag_time="ALAG")
    >>> comp
    Compartment(DEPOT, amount=A_DEPOT(t), lag_time=ALAG)
    >>> dose = Bolus.create("AMT")
    >>> comp = Compartment.create("DEPOT", doses=(dose,))
    >>> comp
    Compartment(DEPOT, amount=A_DEPOT(t), doses=Bolus(AMT, admid=1))
    """

    def __init__(
        self,
        name: str,
        amount: sympy.Expr,
        doses: tuple[Dose, ...] = tuple(),
        input: sympy.Expr = sympy.Integer(0),
        lag_time: sympy.Expr = sympy.Integer(0),
        bioavailability: sympy.Expr = sympy.Integer(1),
    ):
        self._name = name
        self._amount = amount
        self._doses = doses
        self._input = input
        self._lag_time = lag_time
        self._bioavailability = bioavailability

    @classmethod
    def create(
        cls,
        name: str,
        amount: Optional[Union[sympy.Expr, str]] = None,
        doses: Tuple[Dose, ...] = tuple(),
        input: Union[sympy.Expr, str] = sympy.Integer(0),
        lag_time: Union[sympy.Expr, str] = sympy.Integer(0),
        bioavailability: Union[sympy.Expr, str] = sympy.Integer(1),
    ):
        if not isinstance(name, str):
            raise TypeError("Name of a Compartment must be of string type")
        if amount is not None:
            amount = parse_expr(amount)
        else:
            # NOTE: Uses a default idv
            amount = sympy.Function(f'A_{name}')('t')
        if not isinstance(doses, tuple):
            try:
                tuple(doses)
            except TypeError:
                raise TypeError("dose(s) need to be given as a sequence")
        for d in doses:
            if not isinstance(d, Dose):
                raise TypeError("All doses need to be of type Dose")
        input = parse_expr(input)
        lag_time = parse_expr(lag_time)
        bioavailability = parse_expr(bioavailability)
        return cls(
            name=name,
            amount=amount,
            doses=doses,
            input=input,
            lag_time=lag_time,
            bioavailability=bioavailability,
        )

    def replace(self, **kwargs):
        name = kwargs.get("name", self._name)
        amount = kwargs.get("amount", self._amount)
        doses = kwargs.get("doses", self._doses)
        input = kwargs.get("input", self._input)
        lag_time = kwargs.get("lag_time", self._lag_time)
        bioavailability = kwargs.get("bioavailability", self._bioavailability)
        return Compartment.create(
            name=name,
            amount=amount,
            doses=doses,
            input=input,
            lag_time=lag_time,
            bioavailability=bioavailability,
        )

    @property
    def name(self) -> str:
        """Compartment name"""
        return self._name

    @property
    def amount(self) -> sympy.Basic:
        """Compartment amount symbol"""
        return self._amount

    @property
    def doses(self) -> tuple[Dose, ...]:
        # FIXME: Return in defined order with oral doses coming first!
        # Only do this for multiple doses.
        if len(self._doses) > 1:
            return tuple(sorted(self._doses, key=lambda d: isinstance(d, Infusion), reverse=True))
        else:
            return self._doses

    def sorted_doses(self, model):
        """Return doses to compartment where oral doses are located first"""
        if len(self._doses) > 1:
            from pharmpy.modeling.odes import _dose_zo

            return tuple(sorted(self.doses, key=lambda d: _dose_zo(model, d), reverse=True))
        else:
            return self._doses

    @property
    def input(self) -> sympy.Expr:
        return self._input

    @property
    def lag_time(self) -> sympy.Expr:
        """Lag time for doses into compartment"""
        return self._lag_time

    @property
    def bioavailability(self) -> sympy.Expr:
        """Bioavailability fraction for doses into compartment"""
        return self._bioavailability

    @property
    def free_symbols(self) -> set[sympy.Expr]:
        """Get set of all free symbols in the compartment

        Examples
        --------
        >>> from pharmpy.model import Bolus, Compartment
        >>> dose = Bolus.create("AMT")
        >>> comp = Compartment.create("CENTRAL", doses=(dose,), lag_time="ALAG")
        >>> comp.free_symbols  # doctest: +SKIP
        {A_CENTRAL, ALAG, AMT}
        """
        symbs = set()
        for d in self.doses:
            symbs |= d.free_symbols
        symbs |= self.input.free_symbols
        symbs |= self.lag_time.free_symbols
        symbs |= self.bioavailability.free_symbols
        return symbs

    def subs(self, substitutions: Mapping[sympy.Expr, sympy.Expr]) -> Compartment:
        """Substitute expressions or symbols in compartment

        Examples
        --------
        >>> from pharmpy.model import Bolus, Compartment
        >>> dose = Bolus.create("AMT")
        >>> comp = Compartment.create("CENTRAL", doses=(dose,))
        >>> comp.subs({"AMT": "DOSE"})
        Compartment(CENTRAL, amount=A_CENTRAL(t), doses=Bolus(DOSE, admid=1))
        """
        if self.doses:
            new_doses = tuple()
            for d in self.doses:
                new_doses = new_doses + (d.subs(substitutions),)
        else:
            new_doses = tuple()
        return Compartment(
            self.name,
            amount=subs(self._amount, substitutions),
            doses=new_doses,
            input=subs(self._input, substitutions),
            lag_time=subs(self._lag_time, substitutions),
            bioavailability=subs(self._bioavailability, substitutions),
        )

    def __eq__(self, other):
        return (
            isinstance(other, Compartment)
            and self._name == other._name
            and self._amount == other._amount
            and self._doses == other._doses
            and self._input == other._input
            and self._lag_time == other._lag_time
            and self._bioavailability == other._bioavailability
        )

    def __hash__(self):
        return hash(
            (
                self._name,
                self._amount,
                self._doses,
                self._input,
                self._lag_time,
                self._bioavailability,
            )
        )

    def to_dict(self) -> dict[str, Any]:
        if not self._doses:
            all_doses = None
        else:
            all_doses = tuple(dose.to_dict() for dose in self._doses)
        return {
            'class': 'Compartment',
            'name': self._name,
            'amount': sympy.srepr(self._amount),
            'doses': all_doses,
            'input': sympy.srepr(self._input),
            'lag_time': sympy.srepr(self._lag_time),
            'bioavailability': sympy.srepr(self._bioavailability),
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Compartment:
        if d['doses'] is None:
            all_doses = tuple()
        else:
            all_doses = tuple()
            for dose in d['doses']:
                if dose['class'] == 'Bolus':
                    all_doses += (Bolus.from_dict(dose),)
                else:
                    all_doses += (Infusion.from_dict(dose),)
        return cls(
            name=d['name'],
            amount=sympy.parse_expr(d['amount']),
            doses=all_doses,
            input=sympy.parse_expr(d['input']),
            lag_time=sympy.parse_expr(d['lag_time']),
            bioavailability=sympy.parse_expr(d['bioavailability']),
        )

    def __repr__(self):
        lag = '' if self.lag_time == 0 else f', lag_time={self._lag_time}'
        doses = (
            ''
            if self.doses is tuple()
            else f', doses={self._doses[0] if len(self._doses) == 1 else self._doses}'
        )
        input = '' if self.input == 0 else f', input={self._input}'
        bioavailability = (
            '' if self.bioavailability == 1 else f', bioavailability={self._bioavailability}'
        )
        return (
            f'Compartment({self.name}, amount={self._amount}{doses}{input}{lag}{bioavailability})'
        )


class Statements(Sequence, Immutable):
    """A sequence of symbolic statements describing the model

    Two types of statements are supported: Assignment and CompartmentalSystem.
    A Statements object can have 0 or 1 CompartmentalSystem. The order of
    the statements is significant and the same symbol can be assigned
    to multiple times.

    Parameters
    ----------
    statements : list or Statements
        A list of Statement or another Statements to populate this object
    """

    def __init__(self, statements: Optional[Union[Statements, Iterable[Statement]]] = None):
        if isinstance(statements, Statements):
            self._statements = statements._statements
        elif statements is None:
            self._statements = ()
        else:
            self._statements = tuple(statements)

    @overload
    def __getitem__(self, ind: slice) -> Statements:
        ...

    @overload
    def __getitem__(self, ind: int) -> Statement:
        ...

    def __getitem__(self, ind):
        if isinstance(ind, slice):
            return Statements(self._statements[ind])
        else:
            return self._statements[ind]

    def __len__(self):
        return len(self._statements)

    def __add__(self, other: Union[Statements, Statement, Sequence[Statement]]) -> Statements:
        if isinstance(other, Statements):
            return Statements(self._statements + other._statements)
        elif isinstance(other, Statement):
            return Statements(self._statements + (other,))
        else:
            return Statements(self._statements + tuple(other))

    def __radd__(self, other: Union[Statement, Sequence[Statement]]) -> Statements:
        if isinstance(other, Statement):
            return Statements((other,) + self._statements)
        else:
            return Statements(tuple(other) + self._statements)

    @property
    def free_symbols(self) -> set[sympy.Basic]:
        """Get a set of all free symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.free_symbols   # doctest: +SKIP
        {AMT, APGR, A_CENTRAL, BTIME, CL, DV, EPS(1), ETA(1), ETA(2), F, IPRED, IRES, IWRES, S1,
        TAD, THETA(1), THETA(2), THETA(3), TIME, TVCL, TVV, V, W, WGT, Y, t}

        """
        symbols = set()
        for assignment in self:
            symbols |= assignment.free_symbols
        return symbols

    def _get_ode_system_index(self):
        return next(
            map(
                lambda t: t[0],
                filter(lambda t: isinstance(t[1], CompartmentalSystem), enumerate(self)),
            ),
            -1,
        )

    @property
    def ode_system(self) -> Optional[CompartmentalSystem]:
        """Returns the ODE system of the model or None if the model doesn't have an ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.ode_system
        Bolus(AMT, admid=1) → CENTRAL
        ┌───────┐
        │CENTRAL│──CL/V→
        └───────┘
        """
        i = self._get_ode_system_index()
        if i == -1:
            return None
        else:
            cs = self[i]
            assert isinstance(cs, CompartmentalSystem)
            return cs

    @property
    def before_odes(self) -> Statements:
        """All statements before the ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.before_odes
                ⎧TIME  for AMT > 0
                ⎨
        BTIME = ⎩ 0     otherwise
        TAD = -BTIME + TIME
        TVCL = PTVCL⋅WGT
        TVV = PTVV⋅WGT
              ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
              ⎨
        TVV = ⎩       TVV           otherwise
                   ETA₁
        CL = TVCL⋅ℯ
                 ETA₂
        V = TVV⋅ℯ
        S₁ = V
        """
        i = self._get_ode_system_index()
        return self if i == -1 else self[:i]

    @property
    def after_odes(self) -> Statements:
        """All statements after the ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.after_odes
            A_CENTRAL(t)
            ────────────
        F =      S₁
        W = F
        Y = EPS₁⋅W + F
        IPRED = F
        IRES = DV - IPRED
                 IRES
                 ────
        IWRES =   W
        """
        i = self._get_ode_system_index()
        return Statements() if i == -1 else self[i + 1 :]

    @property
    def error(self) -> Statements:
        """All statements after the ODE system or the whole model if no ODE system

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.error
            A_CENTRAL(t)
            ────────────
        F =      S₁
        W = F
        Y = EPS₁⋅W + F
        IPRED = F
        IRES = DV - IPRED
                 IRES
                 ────
        IWRES =   W
        """
        i = self._get_ode_system_index()
        return self if i == -1 else self[i + 1 :]

    def subs(self, substitutions: Mapping[sympy.Basic, sympy.Basic]) -> Statements:
        """Substitute symbols in all statements.

        Parameters
        ----------
        substitutions : dict
            Old-new pairs(can be type str or sympy symbol)

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> stats = model.statements.subs({'WGT': 'WT'})
        >>> stats.before_odes
                ⎧TIME  for AMT > 0
                ⎨
        BTIME = ⎩ 0     otherwise
        TAD = -BTIME + TIME
        TVCL = PTVCL⋅WT
        TVV = PTVV⋅WT
              ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
              ⎨
        TVV = ⎩       TVV           otherwise
                   ETA₁
        CL = TVCL⋅ℯ
                 ETA₂
        V = TVV⋅ℯ
        S₁ = V
        """
        return Statements(s.subs(substitutions) for s in self)

    def _lookup_last_assignment(
        self, symbol: Union[str, sympy.Symbol]
    ) -> Tuple[Optional[int], Optional[Assignment]]:
        if isinstance(symbol, str):
            symbol = sympy.Symbol(symbol)
        ind = None
        assignment = None
        for i, statement in enumerate(self):
            if isinstance(statement, Assignment):
                if statement.symbol == symbol:
                    ind = i
                    assignment = statement
        return ind, assignment

    def find_assignment(self, symbol: Union[str, sympy.Symbol]) -> Optional[Assignment]:
        """Returns last assignment of symbol

        Parameters
        ----------
        symbol : sympy.Symbol or str
            Symbol to look for

        Returns
        -------
        Assignment or None
            An Assignment or None if no assignment to symbol exists

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.find_assignment("CL")
                   ETA₁
        CL = TVCL⋅ℯ
        """
        return self._lookup_last_assignment(symbol)[1]

    def find_assignment_index(self, symbol: Union[str, sympy.Symbol]) -> Optional[int]:
        """Returns index of last assignment of symbol

        Parameters
        ----------
        symbol : sympy.Symbol or str
            Symbol to look for

        Returns
        -------
        int or None
            Index of Assignment or None if no assignment to symbol exists

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.find_assignment_index("CL")
        5
        """
        return self._lookup_last_assignment(symbol)[0]

    def reassign(
        self, symbol: Union[str, sympy.Symbol], expression: Union[str, sympy.Expr]
    ) -> Statements:
        """Reassign symbol to expression

        Set symbol to be expression and remove all previous assignments of symbol

        Parameters
        ----------
        symbol : sypmpy.Symbol or str
            Symbol to reassign
        expression : sympy.Expr or str
            The new expression to assign to symbol

        Return
        ------
        Statements
            Updated statements

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.reassign("CL", "TVCL + eta")   # doctest: +SKIP
        """
        if isinstance(symbol, str):
            symbol = sympy.Symbol(symbol)
        if isinstance(expression, str):
            expression = parse_expr(expression)

        last = True
        new = list(self._statements)
        for i, stat in zip(range(len(new) - 1, -1, -1), reversed(new)):
            if isinstance(stat, Assignment) and stat.symbol == symbol:
                if last:
                    new[i] = Assignment(symbol, expression)
                    last = False
                else:
                    del new[i]
        return Statements(new)

    def _create_dependency_graph(self):
        """Create a graph of dependencies between statements"""
        graph = nx.DiGraph()
        for i in range(len(self) - 1, -1, -1):
            rhs = self[i].rhs_symbols
            for j in range(i - 1, -1, -1):
                statement = self[j]
                if isinstance(statement, Assignment):
                    if statement.symbol in rhs:
                        graph.add_edge(i, j)
                elif isinstance(statement, CompartmentalSystem):
                    amts = {
                        sympy.Function(amt.name)(statement.t) for amt in statement.amounts
                    } | set(statement.amounts)
                    if not rhs.isdisjoint(amts):
                        graph.add_edge(i, j)
        return graph

    def direct_dependencies(self, statement: Statement) -> Statements:
        """Find all direct dependencies of a statement

        Parameters
        ----------
        statement : Statement
            Input statement

        Returns
        -------
        Statements
            Direct dependency statements

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> odes = model.statements.ode_system
        >>> model.statements.direct_dependencies(odes)
                   ETA₁
        CL = TVCL⋅ℯ
                 ETA₂
        V = TVV⋅ℯ
        """
        g = self._create_dependency_graph()
        index = self.index(statement)
        succ = sorted(list(g.successors(index)))
        stats = Statements()
        stats._statements = [self[i] for i in succ]
        return stats

    def dependencies(
        self, symbol_or_statement: Union[str, sympy.Symbol, Statement]
    ) -> set[sympy.Expr]:
        """Find all dependencies of a symbol or statement

        Parameters
        ----------
        symbol : Symbol, str or Statement
            Input symbol or statement

        Returns
        -------
        set
            Set of symbols

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.dependencies("CL")   # doctest: +SKIP
        {ETA(1), THETA(1), WGT}
        """
        if isinstance(symbol_or_statement, Statement):
            i = self.index(symbol_or_statement)
        else:
            symbol = (
                sympy.Symbol(symbol_or_statement)
                if isinstance(symbol_or_statement, str)
                else symbol_or_statement
            )
            for i in range(len(self) - 1, -1, -1):
                statement = self[i]
                if (
                    isinstance(statement, Assignment)
                    and statement.symbol == symbol
                    or isinstance(statement, CompartmentalSystem)
                    and symbol in statement.amounts
                ):
                    break
            else:
                raise KeyError(f"Could not find symbol {symbol}")
        g = self._create_dependency_graph()
        symbs = self[i].rhs_symbols
        if i == 0 or not g:
            # Special case for models with only one statement or no dependent statements
            return symbs
        for j, _ in nx.bfs_predecessors(g, i, sort_neighbors=lambda x: reversed(sorted(x))):
            statement = self[j]
            if isinstance(statement, Assignment):
                symbs -= {statement.symbol}
            else:
                assert isinstance(statement, CompartmentalSystem)
                symbs -= set(statement.amounts)
            symbs |= statement.rhs_symbols
        return symbs

    def remove_symbol_definitions(
        self, symbols: Sequence[sympy.Expr], statement: Statement
    ) -> Statements:
        """Remove symbols and dependencies not used elsewhere

        If the statement no longer depends on the specified
        symbols, this method will make sure that the definitions
        of these symbols will be removed unless they are dependencies
        of other statements.

        Parameters
        ----------
        symbols : iterable
            Iterable of symbols no longer used in the statement
        statement : Statement
            Statement from which the symbols were removed
        """
        graph = self._create_dependency_graph()
        removed_ind = self._statements.index(statement)
        # Statements defining symbols and dependencies
        candidates = set()
        symbols_set = set(symbols)
        for i in range(removed_ind - 1, -1, -1):
            stat = self[i]
            if isinstance(stat, Assignment) and stat.symbol in symbols_set:
                candidates.add(i)
        for i in candidates.copy():
            if i in graph:
                candidates |= set(nx.dfs_preorder_nodes(graph, i))
        # All statements needed for removed_ind
        if removed_ind in graph:
            keep = {down for _, down in nx.dfs_edges(graph, removed_ind)}
        else:
            keep = set()
        candidates -= keep
        # Other dependencies after removed_ind
        additional = {down for up, down in graph.edges if up > removed_ind and down in candidates}
        for add in additional.copy():
            if add in graph:
                additional |= set(nx.dfs_preorder_nodes(graph, add))
        remove = candidates - additional
        return Statements(tuple(self[i] for i in range(len(self)) if i not in remove))

    def full_expression(self, expression: Union[str, sympy.Expr]) -> sympy.Expr:
        """Expand an expression into its full definition

        Parameters
        ----------
        expression : expression or str
            Expression to expand. A string will be converted to an expression.

        Return
        ------
        expression
            Expanded expression

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.statements.before_odes.full_expression("CL")
        PTVCL*WGT*exp(ETA_1)
        """
        if isinstance(expression, str):
            expression = parse_expr(expression)
        for statement in reversed(self):
            if isinstance(statement, CompartmentalSystem):
                raise ValueError(
                    "CompartmentalSystem not supported by full_expression. Use the properties before_odes "
                    "or after_odes."
                )
            expression = subs(
                expression, {statement.symbol: statement.expression}, simultaneous=True
            )
        return expression

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        else:
            for first, second in zip(self, other):
                if first != second:
                    return False
        return True

    def __hash__(self):
        return hash(self._statements)

    def to_dict(self) -> dict[str, Any]:
        stats = tuple(s.to_dict() for s in self)
        return {'statements': stats}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Statements:
        statements = []
        for sdict in d['statements']:
            if sdict['class'] == 'Assignment':
                s = Assignment.from_dict(sdict)
            else:
                s = CompartmentalSystem.from_dict(sdict)
            statements.append(s)
        return cls(tuple(statements))

    def __repr__(self):
        return '\n'.join([repr(statement) for statement in self])

    def _repr_html_(self):
        html = r'\begin{align*}'
        for statement in self:
            if hasattr(statement, '_repr_html_'):
                html += '\\end{align*}'
                s = statement._repr_html_()
                html += s + '\\begin{align*}'
            else:
                s = f'${statement._repr_latex_()}$'
                s = s.replace('=', '&=')
                s = s.replace('$', '')
                s = s + r'\\'
                html += s
        return html + '\\end{align*}'



import pytest
import sympy

from pharmpy.model import (
    Assignment,
    Bolus,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    Infusion,
    Statements,
    output,
)
from pharmpy.modeling import add_effect_compartment, set_first_order_absorption


def S(x):
    return sympy.Symbol(x)


def test_statements_effect_compartment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_pd.mod')
    model = add_effect_compartment(model, "linear")

    with pytest.warns(UserWarning):
        print(model.statements)


def test_str(load_model_for_test, testdata):
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    assert str(s1) == 'KA = X + Y'
    s2 = Assignment(S('X2'), sympy.exp('X'))
    a = str(s2).split('\n')
    assert a[0].startswith(' ')
    assert len(a) == 2

    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert 'TVV' in str(model.statements)
    assert 'TVV' in repr(model.statements)


def test_subs(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements

    s2 = statements.subs({'ETA_1': 'ETAT1'})

    assert s2[5].expression == S('TVCL') * sympy.exp(S('ETAT1'))

    s3 = s2.subs({'TVCL': 'TVCLI'})

    assert s3[2].symbol == S('TVCLI')
    assert s3[5].expression == S('TVCLI') * sympy.exp(S('ETAT1'))

    assert s3.ode_system.free_symbols == {S('V'), S('CL'), S('AMT'), S('t')}

    s4 = s3.subs({'V': 'V2'})

    assert s4.ode_system.free_symbols == {S('CL'), S('AMT'), S('t'), S('V2')}


def test_ode_free_symbols(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    assert model.statements.ode_system.free_symbols == {S('V'), S('CL'), S('AMT'), S('t')}


def test_find_assignment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements

    assert str(statements.find_assignment('CL').expression) == 'TVCL*exp(ETA_1)'
    assert str(statements.find_assignment('S1').expression) == 'V'

    statements = statements + Assignment(S('CL'), S('TVCL') + S('V'))

    assert str(statements.find_assignment('CL').expression) == 'TVCL + V'


def test_eq_assignment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements
    statement_btime = statements[0]
    statement_tad = statements[1]

    assert statement_btime == statement_btime
    assert statement_btime != statement_tad


def test_eq_modelstatements(load_model_for_test, testdata):
    model_min = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    model_pheno = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    assert model_min.statements == model_min.statements
    assert model_pheno.statements == model_pheno.statements
    assert model_min.statements != model_pheno.statements

    s1 = Assignment(S('KA'), S('X') + S('Y'))
    s2 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s3 = Assignment(S('M'), sympy.Integer(2))
    s4 = Assignment(S('G'), sympy.Integer(3))
    s1 = Statements([s2, s1])
    s2 = Statements([s3, s4])
    assert s1 != s2


def test_hash():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    s2 = Assignment(S('KA'), S('X') + S('Z'))
    assert hash(s1) != hash(s2)

    b1 = Bolus(S('AMT'))
    b2 = Bolus(S('AMT'), admid=2)
    assert hash(b1) != hash(b2)

    i1 = Infusion("AMT", rate="R1")
    i2 = Infusion("AMT", rate="R2")
    assert hash(i1) != hash(i2)

    c1 = Compartment.create("DEPOT", lag_time="ALAG")
    c2 = Compartment.create("DEPOT")
    assert hash(c1) != hash(c2)
    assert hash(c1) != hash(output)

    st1 = Statements((s1, s2))
    st2 = Statements((s1,))
    assert hash(st1) != hash(st2)


def test_dict(load_model_for_test, testdata):
    ass1 = Assignment(S('KA'), S('X') + S('Y'))
    d = ass1.to_dict()
    assert d == {
        'class': 'Assignment',
        'symbol': "Symbol('KA')",
        'expression': "Add(Symbol('X'), Symbol('Y'))",
    }
    ass2 = Assignment.from_dict(d)
    assert ass1 == ass2

    dose1 = Bolus.create('AMT')
    d = dose1.to_dict()
    assert d == {'class': 'Bolus', 'amount': "Symbol('AMT')", 'admid': 1}
    dose2 = Bolus.from_dict(d)
    assert dose1 == dose2

    inf1 = Infusion.create('AMT', rate='R1')
    d = inf1.to_dict()
    assert d == {
        'class': 'Infusion',
        'amount': "Symbol('AMT')",
        'admid': 1,
        'rate': "Symbol('R1')",
        'duration': 'None',
    }
    inf2 = Infusion.from_dict(d)
    assert inf1 == inf2

    central = Compartment.create('CENTRAL', doses=(dose1,))
    d = central.to_dict()
    assert d == {
        'class': 'Compartment',
        'name': 'CENTRAL',
        'amount': "Function('A_CENTRAL')(Symbol('t'))",
        'doses': ({'class': 'Bolus', 'amount': "Symbol('AMT')", 'admid': 1},),
        'input': 'Integer(0)',
        'lag_time': 'Integer(0)',
        'bioavailability': 'Integer(1)',
    }
    central2 = Compartment.from_dict(d)
    assert central == central2

    d = output.to_dict()
    assert d == {'class': 'Output'}
    output2 = output.__class__.from_dict(d)
    assert output == output2

    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    odes = model.statements.ode_system
    d = odes.to_dict()
    assert d == {
        'class': 'CompartmentalSystem',
        'compartments': (
            {'class': 'Output'},
            {
                'class': 'Compartment',
                'name': 'CENTRAL',
                'amount': "Function('A_CENTRAL')(Symbol('t'))",
                'doses': ({'class': 'Bolus', 'amount': "Symbol('AMT')", 'admid': 1},),
                'input': 'Integer(0)',
                'lag_time': 'Integer(0)',
                'bioavailability': 'Integer(1)',
            },
        ),
        'rates': [(1, 0, "Mul(Symbol('CL'), Pow(Symbol('V'), Integer(-1)))")],
        't': "Symbol('t')",
    }
    odes2 = CompartmentalSystem.from_dict(d)
    assert odes == odes2

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    d = model.statements.to_dict()
    assert d == {
        'statements': (
            {
                'class': 'Assignment',
                'symbol': "Symbol('Y')",
                'expression': "Add(Symbol('EPS_1'), Symbol('ETA_1'), Symbol('THETA_1'))",
            },
        )
    }
    stats2 = Statements.from_dict(d)
    assert model.statements == stats2


def test_add():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    s2 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s3 = Assignment(S('M'), sympy.Integer(2))
    s = Statements([s1, s2])
    new = (s3,) + s
    assert len(new) == 3


def test_remove_symbol_definition():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    s2 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s3 = Assignment(S('M'), sympy.Integer(2))
    s4 = Assignment(S('G'), sympy.Integer(3))
    s = Statements([s4, s3, s2, s1])
    ns = s.remove_symbol_definitions([S('Z')], s1)
    assert ns == Statements([s4, s1])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X') + S('K'))
    s = Statements([s1, s2, s3, s4])
    ns = s.remove_symbol_definitions([S('CL')], s4)
    assert ns == Statements([s1, s4])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X') + S('K'))
    s5 = Assignment(S('KR'), S('CL'))
    s = Statements([s1, s2, s3, s4, s5])
    ns = s.remove_symbol_definitions([S('CL')], s4)
    assert ns == Statements([s1, s2, s3, s4, s5])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X'))
    s = Statements([s1, s2, s3, s4])
    ns = s.remove_symbol_definitions([S('CL'), S('K')], s4)
    assert ns == Statements([s4])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X'))
    s5 = Assignment(S('P'), S('K'))
    s = Statements([s1, s2, s3, s4, s5])
    ns = s.remove_symbol_definitions([S('CL'), S('K')], s4)
    assert ns == Statements([s1, s4, s5])


def test_reassign():
    s1 = Assignment(S('G'), sympy.Integer(3))
    s2 = Assignment(S('M'), sympy.Integer(2))
    s3 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s4 = Assignment(S('KA'), S('X') + S('Y'))
    s = Statements([s1, s2, s3, s4])
    snew = s.reassign(S('M'), S('x') + S('y'))
    assert snew == Statements([s1, Assignment(S('M'), S('x') + S('y')), s3, s4])

    s5 = Assignment(S('KA'), S('KA') + S('Q') + 1)
    s = Statements([s1, s2, s3, s4, s5])
    snew = s.reassign(S('KA'), S('F'))
    assert snew == Statements([s1, s2, s3, Assignment(S('KA'), S('F'))])
    snew = s.reassign('KA', 'F')
    assert snew == Statements([s1, s2, s3, Assignment(S('KA'), S('F'))])


def test_find_compartment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    comp = model.statements.ode_system.find_compartment('CENTRAL')
    assert comp.name == 'CENTRAL'
    comp = model.statements.ode_system.find_compartment('NOTINMODEL')
    assert comp is None


def test_dosing_compartment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert model.statements.ode_system.dosing_compartments[0].name == 'CENTRAL'


def test_central_compartment(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.central_compartment.name == 'CENTRAL'
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_nodepot.mod')
    assert model.statements.ode_system.central_compartment.name == 'CENTRAL'


def test_find_depot(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.find_depot(model.statements).name == 'DEPOT'
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod')
    assert model.statements.ode_system.find_depot(model.statements) is None
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_depot.mod')
    assert model.statements.ode_system.find_depot(model.statements).name == 'DEPOT'
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_nodepot.mod')
    assert model.statements.ode_system.find_depot(model.statements) is None


def test_peripheral_compartments(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.find_peripheral_compartments() == []


def test_find_transit_compartments(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod')
    assert model.statements.ode_system.find_transit_compartments(model.statements) == []
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.find_transit_compartments(model.statements) == []
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_1transit.mod')
    transits = model.statements.ode_system.find_transit_compartments(model.statements)
    assert len(transits) == 1
    assert transits[0].name == 'TRANS1'
    model = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_2transits.mod')
    transits = model.statements.ode_system.find_transit_compartments(model.statements)
    assert len(transits) == 2
    assert transits[0].name == 'TRANS1'
    assert transits[1].name == 'TRANS2'


def test_before_odes(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    before_ode = model.statements.before_odes
    assert before_ode[-1].symbol.name == 'S1'


def test_full_expression(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    expr = model.statements.before_odes.full_expression("CL")
    assert expr == sympy.Symbol("PTVCL") * sympy.Symbol("WGT") * sympy.exp(sympy.Symbol("ETA_1"))
    with pytest.raises(ValueError):
        model.statements.full_expression("Y")


def test_to_explicit_ode_system(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    cs = model.statements.ode_system
    assert len(cs.eqs) == 1

    assert cs.amounts == sympy.Matrix([sympy.Function('A_CENTRAL')('t')])


def test_repr_latex():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    latex = s1._repr_latex_()
    assert latex == r'$KA = X + Y$'


def test_repr_html():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    stats = Statements([s1])
    html = stats._repr_html_()
    assert 'X + Y' in html

    cb = CompartmentalSystemBuilder()
    dose = Bolus.create('AMT')
    central = Compartment.create('CENTRAL', doses=(dose,))
    cb.add_compartment(central)
    cb.add_compartment(output)
    cb.add_flow(central, output, S('K'))
    cs = CompartmentalSystem(cb)
    stats = Statements([cs])
    assert type(stats._repr_html_()) == str


def test_direct_dependencies(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    odes = model.statements.ode_system
    deps = model.statements.direct_dependencies(odes)
    assert deps[0].symbol.name == "CL"
    assert deps[1].symbol.name == "V"


def test_dependencies(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    depsy = model.statements.dependencies(S('Y'))
    assert depsy == {
        S('EPS_1'),
        S('AMT'),
        S('PTVCL'),
        S('t'),
        S('PTVV'),
        S('THETA_3'),
        S('APGR'),
        S('WGT'),
        S('ETA_2'),
        S('ETA_1'),
    }
    depscl = model.statements.dependencies(S('CL'))
    assert depscl == {S('PTVCL'), S('WGT'), S('ETA_1')}
    odes = model.statements.ode_system
    deps_odes = model.statements.dependencies(odes)
    assert deps_odes == {
        S('AMT'),
        S('APGR'),
        S('ETA_1'),
        S('ETA_2'),
        S('PTVCL'),
        S('PTVV'),
        S('THETA_3'),
        S('WGT'),
        S('t'),
    }
    with pytest.raises(KeyError):
        model.statements.dependencies("NONEXISTING")


def test_builder():
    cb = CompartmentalSystemBuilder()
    dose = Bolus('AMT')
    central = Compartment.create('CENTRAL', doses=(dose,))
    cb.add_compartment(central)
    cb.add_flow(central, output, S('K'))
    depot = Compartment.create('DEPOT')
    cb.add_compartment(depot)
    cb.add_flow(depot, central, S('KA'))
    cm = CompartmentalSystem(cb)
    assert not cm.find_compartment('DEPOT').doses
    assert cm.central_compartment.doses[0] == dose
    cb.move_dose(central, depot)
    cm2 = CompartmentalSystem(cb)
    assert cm2.find_compartment('DEPOT').doses[0] == dose
    assert not cm2.central_compartment.doses
    assert hash(cm) != hash(cm2)


def test_infusion_repr():
    inf = Infusion.create('AMT', rate='R1')
    assert repr(inf) == 'Infusion(AMT, admid=1, rate=R1)'
    inf = Infusion.create('AMT', duration='D1')
    assert repr(inf) == 'Infusion(AMT, admid=1, duration=D1)'
    inf = Infusion.create('AMT', admid=2, duration='D1')
    assert repr(inf) == 'Infusion(AMT, admid=2, duration=D1)'


def test_infusion_create():
    inf = Infusion.create('AMT', rate='R1')
    assert inf.rate == S('R1')

    with pytest.raises(ValueError):
        Infusion.create('AMT', rate='R1', duration='D1')

    with pytest.raises(ValueError):
        Infusion.create('AMT')


def test_compartment_repr():
    comp = Compartment.create("CENTRAL", lag_time='LT')
    assert repr(comp) == "Compartment(CENTRAL, amount=A_CENTRAL(t), lag_time=LT)"


def test_compartment_names(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert model.statements.ode_system.compartment_names == ['CENTRAL']


def test_assignment_create_numeric(load_model_for_test, testdata):
    with pytest.raises(AttributeError):
        Assignment('X', 1).free_symbols
    assert Assignment.create('X', 1).free_symbols
    with pytest.raises(AttributeError):
        Assignment('X', 1.0).free_symbols
    assert Assignment.create('X', 1.0).free_symbols


def test_multi_dose_comp_order(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_first_order_absorption(model)

    ode = model.statements.ode_system
    cb = CompartmentalSystemBuilder(ode)
    cb.set_dose(cb.find_compartment("CENTRAL"), cb.find_compartment("DEPOT").doses)
    ode = CompartmentalSystem(cb)
    assert ode.dosing_compartments[0].name == "DEPOT"



import shutil

import pandas as pd
import pytest

import pharmpy.modeling as modeling
from pharmpy.config import site_config_path, user_config_path
from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import fit
from pharmpy.tools.external.nlmixr import verification as nlmixr_verification
from pharmpy.tools.external.nonmem import conf
from pharmpy.tools.external.nonmem.run import execute_model as nonmem_execute_model
from pharmpy.tools.external.rxode import verification as rxode_verification
from pharmpy.workflows import LocalDirectoryToolDatabase, ModelEntry


def test_configuration():
    print("User config dir:", user_config_path())
    print("Site config dir:", site_config_path())
    print("Default NONMEM path:", conf.default_nonmem_path)
    assert (conf.default_nonmem_path / 'license' / 'nonmem.lic').is_file()


def test_fit_single(tmp_path, model_count, testdata):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model.parse_model('pheno.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        res = fit(model)
        rundir = tmp_path / 'modelfit_dir1'
        assert res.ofv == pytest.approx(730.8947268137308)
        assert rundir.is_dir()
        assert model_count(rundir) == 1
        assert (rundir / 'models' / 'pheno' / '.pharmpy').exists()
        assert not [
            path.name for path in (rundir / 'models' / 'pheno').iterdir() if 'contr' in path.name
        ]


def test_fit_multiple(tmp_path, model_count, testdata):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno_1.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno_1.dta')
        model_1 = Model.parse_model('pheno_1.mod')
        df = pd.read_table(tmp_path / 'pheno_1.dta', sep=r'\s+', header=0)
        model_1 = model_1.replace(
            dataset=df, datainfo=model_1.datainfo.replace(path=tmp_path / 'pheno_1.dta')
        )
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno_2.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno_2.dta')
        model_2 = Model.parse_model('pheno_2.mod')
        model_2 = model_2.replace(
            dataset=df, datainfo=model_2.datainfo.replace(path=tmp_path / 'pheno_2.dta')
        )
        res1, res2 = fit([model_1, model_2])
        rundir = tmp_path / 'modelfit_dir1'
        assert res1.ofv == pytest.approx(730.8947268137308)
        assert res2.ofv == pytest.approx(730.8947268137308)
        assert rundir.is_dir()
        assert model_count(rundir) == 2


def test_fit_copy(tmp_path, model_count, testdata):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path / 'pheno.mod')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path / 'pheno.dta')

        model_1 = Model.parse_model('pheno.mod')
        model_1 = model_1.replace(datainfo=model_1.datainfo.replace(path=tmp_path / 'pheno.dta'))
        res1 = fit(model_1)

        rundir_1 = tmp_path / 'modelfit_dir1'
        assert rundir_1.is_dir()
        assert model_count(rundir_1) == 1

        model_2 = model_1.replace(name='pheno_copy')
        model_2 = modeling.update_inits(model_2, res1.parameter_estimates)
        res2 = fit(model_2)

        rundir_2 = tmp_path / 'modelfit_dir1'
        assert rundir_2.is_dir()
        assert model_count(rundir_2) == 1

        assert res1.ofv != res2.ofv


def test_fit_nlmixr(tmp_path, testdata):
    from pharmpy.tools.external.nlmixr import conf

    if str(conf.rpath) == '.':
        pytest.skip("No R selected in conf. Skipping nlmixr tests")
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model.parse_model('pheno.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        model = modeling.convert_model(model, 'nlmixr')
        res = fit(model, tool='nlmixr')
        assert res.ofv == pytest.approx(732.58813)
        assert res.parameter_estimates['TVCL'] == pytest.approx(0.0058686, abs=1e-6)


def test_verification_nlmixr(tmp_path, testdata):
    from pharmpy.tools.external.nlmixr import conf

    if str(conf.rpath) == '.':
        pytest.skip("No R selected in conf. Skipping nlmixr tests")
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_real.ctl')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model.parse_model('pheno_real.ctl')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        assert nlmixr_verification(model)


def test_verification_rxode(tmp_path, testdata):
    from pharmpy.tools.external.nlmixr import conf

    if str(conf.rpath) == '.':
        pytest.skip("No R selected in conf. Skipping rxode tests")
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_real.ctl')
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model = Model.parse_model('pheno_real.ctl')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        assert rxode_verification(model)


def test_execute_model_nonmem(tmp_path, testdata):
    with chdir(tmp_path):
        datadir = testdata / 'nonmem'
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy(datadir / 'pheno.dta', 'pheno.dta')
        model = Model.parse_model('pheno_real.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))

        model_entry = ModelEntry.create(
            model=model,
            modelfit_results=None,
            parent=model.parent_model,
            log=None,
        )

        db = LocalDirectoryToolDatabase('db_model')

        model_entry = nonmem_execute_model(model_entry, db)

        assert isinstance(model_entry, ModelEntry)
        assert model_entry.modelfit_results
        assert not model_entry.model.modelfit_results
        assert (db.path / 'models' / 'pheno_real').is_dir()
